from scarabee import (
    NoTarget,
    SingleTarget,
    Branch,
    BranchingTargets,
    FissionYields,
    ChainEntry,
    DepletionChain,
    NDLibrary,
)
import numpy as np
import xml.etree.ElementTree as ET
from typing import Union


def strip_name(name) -> str:
    if "_" in name:
        return name.replace("_", "")
    return name


def read_fission_yields(node) -> Union[FissionYields, str]:
    if "parent" in node.attrib:
        # raise RuntimeError("Fission yields refers to parent nuclide {:}.".format(node.attrib['parent']))
        return strip_name(node.attrib["parent"])

    energies = node[0].text.split()
    energies = [float(E) for E in energies]
    NE = len(energies)  # Number of energies

    products = None
    yields = []

    for i in range(1, NE + 1):
        fiss_yields = node[i]

        prods = fiss_yields[0].text.split()
        prods = [strip_name(n) for n in prods]
        if products is None:
            products = prods
        else:
            # Make sure products are in the same order !
            if len(products) != len(prods):
                raise RuntimeError(
                    "Product lists at different energies are of different lengths."
                )
            for i in range(len(products)):
                if products[i] != prods[i]:
                    raise RuntimeError(
                        "Product lists at different energies do not agree."
                    )

        i_yields = fiss_yields[1].text.split()
        yields.append([float(y) for y in i_yields])

    yields = np.array(yields)

    return FissionYields(products, energies, yields)


def build_depletion_chain(omc_fname) -> DepletionChain:
    root = ET.parse(omc_fname).getroot()

    dc = DepletionChain()

    # First, go through all nuclides and obtain all fission yields which don't
    # just refer to a different nuclide.
    fiss_yields = {}
    for nuc in root:
        if (
            len(nuc) > 0
            and nuc[-1].tag == "neutron_fission_yields"
            and "parent" not in nuc[-1].attrib
        ):
            nuclide_name = strip_name(nuc.attrib["name"])
            fiss_yields[nuclide_name] = read_fission_yields(nuc[-1])

    # Parse all nuclides in the chain
    for nuc in root:
        nuclide_name = strip_name(nuc.attrib["name"])

        half_life = None
        decay_branches_list = []

        n_gamma = None
        n_2n = None
        n_3n = None
        n_p = None
        n_alpha = None
        n_fission = None

        if "half_life" in nuc.attrib:
            half_life = float(nuc.attrib["half_life"])

        for entry in nuc:
            if entry.tag == "decay":
                if "target" in entry.attrib:
                    branch = Branch()
                    branch.target = strip_name(entry.attrib["target"])
                    branch.branch_ratio = float(entry.attrib["branching_ratio"])
                    if branch.branch_ratio > 0.0:
                        decay_branches_list.append(branch)

            elif entry.tag == "reaction":
                reaction = entry.attrib["type"]
                transmut_target = None
                branch_ratio = None
                if "target" in entry.attrib:
                    transmut_target = strip_name(entry.attrib["target"])
                if "branching_ratio" in entry.attrib:
                    branch_ratio = float(entry.attrib["branching_ratio"])

                if reaction == "(n,gamma)":
                    if branch_ratio is None and transmut_target is None:
                        n_gamma = NoTarget()
                    elif branch_ratio is None:
                        n_gamma = SingleTarget(transmut_target)
                    else:
                        if n_gamma is None:
                            n_gamma = []
                        branch = Branch()
                        branch.target = transmut_target
                        branch.branch_ratio = branch_ratio
                        n_gamma.append(branch)

                elif reaction == "(n,2n)":
                    if branch_ratio is None and transmut_target is None:
                        n_2n = NoTarget()
                    elif branch_ratio is None:
                        n_2n = SingleTarget(transmut_target)
                    else:
                        if n_2n is None:
                            n_2n = []
                        branch = Branch()
                        branch.target = transmut_target
                        branch.branch_ratio = branch_ratio
                        n_2n.append(branch)

                elif reaction == "(n,3n)":
                    if branch_ratio is None and transmut_target is None:
                        n_3n = NoTarget()
                    elif branch_ratio is None:
                        n_3n = SingleTarget(transmut_target)
                    else:
                        if n_3n is None:
                            n_3n = []
                        branch = Branch()
                        branch.target = transmut_target
                        branch.branch_ratio = branch_ratio
                        n_3n.append(branch)

                elif reaction == "(n,p)":
                    if branch_ratio is None and transmut_target is None:
                        n_p = NoTarget()
                    elif branch_ratio is None:
                        n_p = SingleTarget(transmut_target)
                    else:
                        if n_p is None:
                            n_p = []
                        branch = Branch()
                        branch.target = transmut_target
                        branch.branch_ratio = branch_ratio
                        n_p.append(branch)

                elif reaction == "(n,a)":
                    if branch_ratio is None and transmut_target is None:
                        n_alpha = NoTarget()
                    elif branch_ratio is None:
                        n_alpha = SingleTarget(transmut_target)
                    else:
                        if n_alpha is None:
                            n_alpha = []
                        branch = Branch()
                        branch.target = transmut_target
                        branch.branch_ratio = branch_ratio
                        n_alpha.append(branch)
                else:
                    # OpenMC considers others like (n,4n), but Scarabée doesn't
                    # track these other transmutation reactions.
                    pass

            elif entry.tag == "neutron_fission_yields":
                parent = nuclide_name
                if "parent" in entry.attrib:
                    parent = strip_name(entry.attrib["parent"])
                n_fission = fiss_yields[parent]

        # Create decay targets (if there are any)
        decay_targets = None
        if len(decay_branches_list) == 1:
            decay_targets = SingleTarget(decay_branches_list[0].target)
        elif len(decay_branches_list) > 1:
            decay_targets = BranchingTargets(decay_branches_list)
        if half_life is not None and decay_targets is None:
            decay_targets = NoTarget()

        if isinstance(n_gamma, list):
            n_gamma = BranchingTargets(n_gamma)
        if isinstance(n_2n, list):
            n_2n = BranchingTargets(n_2n)
        if isinstance(n_3n, list):
            n_3n = BranchingTargets(n_3n)
        if isinstance(n_p, list):
            n_p = BranchingTargets(n_p)
        if isinstance(n_alpha, list):
            n_alpha = BranchingTargets(n_alpha)

        # Build the chain entry for the nuclide
        ce = ChainEntry()
        ce.half_life = half_life
        ce.decay_targets = decay_targets
        ce.n_gamma = n_gamma
        ce.n_2n = n_2n
        ce.n_3n = n_3n
        ce.n_p = n_p
        ce.n_alpha = n_alpha
        ce.n_fission = n_fission

        dc.insert_entry(nuclide_name, ce)

    # Expunge all short-lives nuclides (i.e. half life less than 24 hrs) that
    # aren't in important chains.
    nuclide_names = dc.nuclides
    for nuc_name in nuclide_names:
        if nuc_name in ["I135", "Xe135"]:
            continue

        nuc = dc.nuclide_data(nuc_name)

        if nuc.half_life is not None and nuc.half_life < 60.0 * 60.0 * 24.0:
            print("Removing {:} from chain".format(nuc_name))
            dc.remove_nuclide(nuc_name)
        elif nuc_name in [
            "Cd115",
            "Rh102",
            "Rh102m1",
            "Sb127",
            "Br82",
        ]:  # No Evals in ENDF-8.0 for these. RIP
            print("Removing {:} from chain".format(nuc_name))
            dc.remove_nuclide(nuc_name)

    return dc


def fix_target_name(name) -> str:
    if name[-2:] == "m1":
        name = name[:-2] + "_m1"
    return name


def print_no_target(name: str, target: NoTarget) -> None:
    print(f'    <reaction type="{name}" Q="0.0" />')


def print_single_target(name: str, target: SingleTarget) -> None:
    target_name = target.target
    target_name = fix_target_name(target_name)
    print(f'    <reaction type="{name}" Q="0.0" target="{target_name}" />')


def print_branching_targets(name: str, target: BranchingTargets) -> None:
    # OpenMC can have a branching reaction target. In this case, we find the largets
    max = 0.0
    max_i = -1
    for i in range(len(target.branches)):
        if target.branches[i].branch_ratio > max:
            max = target.branches[i].branch_ratio
            max_i = i

    branch = target.branches[max_i]
    target_name = branch.target
    target_name = fix_target_name(target_name)
    # print(f'    <reaction type="{name}" Q="0.0" branching_ratio="{branch.branch_ratio}" target="{target_name}"/>')
    print(f'    <reaction type="{name}" Q="0.0" target="{target_name}" />')


def print_fission_yields(
    name: str, nuc: str, target: FissionYields, ndl: NDLibrary
) -> None:
    nucdata = ndl.get_nuclide(nuc)
    Q = nucdata.fission_energy * 1.0e6

    print(f'    <reaction type="{name}" Q="{Q}" />')

    # Now must print fission yields !
    print("    <neutron_fission_yields>")

    # Print list of tabulated energies
    energies_str = "<energies>"
    for E in target.incident_energies:
        energies_str += f"{E:.8E} "
    energies_str = energies_str.strip()
    energies_str = "      " + energies_str + "</energies>"
    print(energies_str)

    # Print yields for each tabulated energy
    for i, E in enumerate(target.incident_energies):
        print(f'      <fission_yields energy="{E}">')

        targets = target.targets
        for i in range(len(targets)):
            targets[i] = fix_target_name(targets[i])

        yields = [0.0 for t in range(len(targets))]
        for i in range(len(targets)):
            yields[i] = target.fission_yield(i, E)

        targets_str = "<products>"
        for tr in targets:
            targets_str += f"{tr} "
        targets_str = targets_str.strip()
        targets_str = "        " + targets_str + "</products>"
        print(targets_str)

        yields_str = "<data>"
        for y in yields:
            yields_str += f"{y:.6E} "
        yields_str = yields_str.strip()
        yields_str = "        " + yields_str + "</data>"
        print(yields_str)

        print(f"      </fission_yields>")

    print("    </neutron_fission_yields>")


def print_reaction(
    name: str,
    nuc: str,
    target: Union[NoTarget, SingleTarget, BranchingTargets, FissionYields],
    ndl: NDLibrary,
) -> None:
    if isinstance(target, NoTarget):
        print_no_target(name, target)
    elif isinstance(target, SingleTarget):
        print_single_target(name, target)
    elif isinstance(target, BranchingTargets):
        print_branching_targets(name, target)
    elif isinstance(target, FissionYields):
        print_fission_yields(name, nuc, target, ndl)
    else:
        raise RuntimeError("Should never get here !")


def print_nuclide(nucname: str, entry: ChainEntry, ndl: NDLibrary) -> None:
    # Compute number of reactions
    NR = 0
    if entry.n_gamma:
        NR += 1
    if entry.n_2n:
        NR += 1
    if entry.n_3n:
        NR += 1
    if entry.n_p:
        NR += 1
    if entry.n_alpha:
        NR += 1
    if entry.n_fission:
        NR += 1

    # Get number of decay modes and number of reactions
    ND = 0
    if entry.half_life:
        if isinstance(entry.decay_targets, NoTarget):
            ND = 1
        elif isinstance(entry.decay_targets, SingleTarget):
            ND = 1
        elif isinstance(entry.decay_targets, BranchingTargets):
            ND = len(entry.decay_targets.branches)
        else:
            raise RuntimeError("Should never get here !")

    # Get half-life
    HL = 0.0
    if ND != 0:
        HL = entry.half_life

    # Print nuclide header
    if ND != 0:
        print(
            f'  <nuclide name="{fix_target_name(nucname)}" decay_modes="{ND}" reactions="{NR}" half_life="{HL:.6E}">'
        )
    else:
        print(
            f'  <nuclide name="{fix_target_name(nucname)}" decay_modes="{ND}" reactions="{NR}">'
        )

    # Print all decays (just say they are all beta- or something)
    if entry.decay_targets is not None:
        if isinstance(entry.decay_targets, NoTarget):
            print(f'    <decay type="beta" branching_ratio="1.0" />')
        elif isinstance(entry.decay_targets, SingleTarget):
            target_name = entry.decay_targets.target
            target_name = fix_target_name(target_name)
            print(
                f'    <decay type="beta" branching_ratio="1.0" target="{target_name}" />'
            )
        elif isinstance(entry.decay_targets, BranchingTargets):
            for i in range(len(entry.decay_targets.branches)):
                branch = entry.decay_targets.branches[i]
                target_name = branch.target
                target_name = fix_target_name(target_name)
                print(
                    f'    <decay type="beta" branching_ratio="{branch.branch_ratio}" target="{target_name}" />'
                )
        else:
            raise RuntimeError("Should never get here !")

    # Print reactions
    if entry.n_gamma:
        print_reaction("(n,gamma)", nucname, entry.n_gamma, ndl)
    if entry.n_2n:
        print_reaction("(n,2n)", nucname, entry.n_2n, ndl)
    if entry.n_3n:
        print_reaction("(n,3n)", nucname, entry.n_3n, ndl)
    if entry.n_p:
        print_reaction("(n,p)", nucname, entry.n_p, ndl)
    if entry.n_alpha:
        print_reaction("(n,a)", nucname, entry.n_alpha, ndl)
    if entry.n_fission:
        print_reaction("fission", nucname, entry.n_fission, ndl)

    # Close nuclide
    print("  </nuclide>")


def print_chain(chain: DepletionChain, ndl: NDLibrary) -> None:
    print('<?xml version="1.0"?>')
    print("<depletion_chain>")

    # List of all nuclides
    nuclides = list(chain.nuclides)

    for nuc in nuclides:
        print_nuclide(nuc, chain.nuclide_data(nuc), ndl)

    print("</depletion_chain>")



from enum import Enum

class CriticalLeakage(Enum):
  """
  Defines a critical leakage model which can be applied to an assembly.
  """

  P1 = 1
  """
  Homogeneous P1 approximation.
  """

  B1 = 2
  """
  Homogeneous B1 approximation.
  """

  NoLeakage = 3
  """
  No leakage (assembly calculation is not modified).
  """



# Analytical Benchmark Test Set for Criticality Code and Verification
# Problem No.:32 [PUa-1-1-SL]
# Exact keff = 1.

from scarabee import *
import numpy as np
import matplotlib.pyplot as plt

Et = np.array([1.])
Ea = np.array([1. - 0.733333])
Es = np.array([[[0.733333]],
              [[0.333333]]])
Ef = np.array([0.266667])
vEf = np.array([2.5 * 0.266667])
chi = np.array([1.])

dtr = np.array([0.333333])

fuel = CrossSection(Et, dtr, Ea, Es, Ef, vEf, chi)

# Define Cells

thickness =  0.79606 * 2.
Nx = 500
# slab
dx = thickness / Nx
dy = thickness
slab = EmptyCell(fuel, dx, dy)
fuel_slab = Cartesian2D(Nx*[dx], [dy])
fuel_slab.set_tiles([slab] * Nx)

moc = MOCDriver(fuel_slab, anisotropic = True)
moc.x_min_bc = BoundaryCondition.Vacuum
moc.x_max_bc = BoundaryCondition.Vacuum
moc.y_min_bc = BoundaryCondition.Reflective
moc.y_max_bc = BoundaryCondition.Reflective

moc.generate_tracks(128, 0.005, YamamotoTabuchi6())
#moc.generate_tracks(128, 0.01, Legendre12())
moc.flux_tolerance = 1.E-5
moc.keff_tolerance = 1.E-5

moc.solve()

flux, x, y = moc.rasterize_flux(1000, 1)
plt.stairs(flux[0,0,:], edges=x)
plt.show()



from scarabee import (
    NDLibrary,
    MaterialComposition,
    Material,
    Fraction,
    DensityUnits,
    set_output_file,
)
from scarabee.reseau import FuelPin, GuideTube, PWRAssembly, Symmetry

name = "F24_0"

set_output_file(name + "_out.txt")

ndl = NDLibrary()

# Define all Materials
Fuel24Comp = MaterialComposition(Fraction.Atoms, name="Fuel 2.4%")
Fuel24Comp.add_leu(2.4, 1.0)
Fuel24Comp.add_element("O", 2.0)
Fuel24 = Material(Fuel24Comp, 575.0, 10.29748, DensityUnits.g_cm3, ndl)

CladComp = MaterialComposition(Fraction.Weight, name="Zircaloy 4")
CladComp.add_element("O", 0.00125)
CladComp.add_element("Cr", 0.0010)
CladComp.add_element("Fe", 0.0021)
CladComp.add_element("Zr", 0.98115)
CladComp.add_element("Sn", 0.0145)
Clad = Material(CladComp, 575.0, 6.55, DensityUnits.g_cm3, ndl)

HeComp = MaterialComposition(Fraction.Atoms, name="He Gas")
HeComp.add_element("He", 1.0)
He = Material(HeComp, 575.0, 0.0015981, DensityUnits.g_cm3, ndl)

# Define a guide tube
gt = GuideTube(inner_radius=0.56134, outer_radius=0.60198, clad=Clad)

# Define fuel pin
fp = FuelPin(
    fuel=Fuel24,
    fuel_radius=0.39218,
    gap=He,
    gap_radius=0.40005,
    clad=Clad,
    clad_radius=0.45720,
)

cells = [
    [fp, fp, fp, fp, fp, fp, fp, fp, fp],
    [fp, fp, fp, fp, fp, fp, fp, fp, fp],
    [gt, fp, fp, gt, fp, fp, fp, fp, fp],
    [fp, fp, fp, fp, fp, gt, fp, fp, fp],
    [fp, fp, fp, fp, fp, fp, fp, fp, fp],
    [gt, fp, fp, gt, fp, fp, gt, fp, fp],
    [fp, fp, fp, fp, fp, fp, fp, fp, fp],
    [fp, fp, fp, fp, fp, fp, fp, fp, fp],
    [gt, fp, fp, gt, fp, fp, gt, fp, fp],
]

# Define assembly
asmbly = PWRAssembly(
    pitch=1.25984,
    assembly_pitch=21.50364,
    shape=(17, 17),
    symmetry=Symmetry.Quarter,
    moderator_pressure=15.5132,
    moderator_temp=575.0,
    boron_ppm=975.0,
    cells=cells,
    ndl=ndl,
)

asmbly.solve()
asmbly.diffusion_data.save(name + ".bin")



# Analytical Benchmark Test Set for Criticality Code and Verification
# Problem No.:32 [PUa-1-1-SL]
# Exact keff = 1.

from scarabee import *
import numpy as np
import matplotlib.pyplot as plt

Et = np.array([1.])
Ea = np.array([1. - 0.733333])
Es = np.array([[[0.733333]],
              [[0.333333]]])
Ef = np.array([0.266667])
vEf = np.array([2.5 * 0.266667])
chi = np.array([1.])

dtr = np.array([0.333333])

fuel = CrossSection(Et, dtr, Ea, Es, Ef, vEf, chi)

# Define Cells

thickness =  0.79606
Nx = 5000
# slab
dx = thickness / Nx

sn = ReflectorSN(Nx*[fuel], Nx*[dx], 128, True)
print(sn.anisotropic, sn.max_legendre_order)
sn.solve()

x = np.arange(start=0., stop=thickness, step=dx)
flux = np.zeros(Nx)
for i in range(Nx):
  flux[i] = sn.flux(i, 0)

plt.plot(x, flux)
plt.show()



from scarabee import *
import numpy as np
import matplotlib.pyplot as plt

ndl = NDLibrary()

UO2comp = MaterialComposition()
UO2comp.add_nuclide("U235", 7.0803E-4)
UO2comp.add_nuclide("U238", 2.2604E-2)
UO2comp.add_nuclide("O16",  4.6624E-2)
UO2 = Material(UO2comp, 293.6, ndl)

Zirccomp = MaterialComposition()
Zirccomp.add_nuclide("Zr90", 2.2200E-2)
Zirccomp.add_nuclide("Zr91", 4.8280E-3)
Zirccomp.add_nuclide("Zr92", 7.3713E-3)
Zirccomp.add_nuclide("Zr94", 7.5006E-3)
Zirccomp.add_nuclide("Zr96", 1.2070E-3)
Zirc = Material(Zirccomp, 293.6, ndl)

Watercomp = MaterialComposition()
Watercomp.add_nuclide("H1_H2O", 6.6630E-2)
Watercomp.add_nuclide("O16",    3.3315E-2)
Water = Material(Watercomp, 293.6, ndl)

Et = np.array([1.E5])
Ea = np.array([1.E5])
Es = np.array([[0.]])
Fuel = CrossSection(Et, Ea, Es, "Fuel")

Et = np.array([Zirc.potential_xs])
Ea = np.array([Zirc.potential_xs])
Clad = CrossSection(Et, Ea, Es, "Clad")

Et = np.array([Water.potential_xs])
Ea = np.array([Water.potential_xs])
Mod = CrossSection(Et, Ea, Es, "Mod")

# Define Cells
pitch = 1.26
iso_pitch = 500.*pitch
fuel_rad = 0.4095
clad_rad = 0.475

# Simulate isolated geometry
radii = [fuel_rad, clad_rad]
mats =  [Fuel, Clad, Mod]
isolated_pin = SimplePinCell(radii, mats, iso_pitch, iso_pitch)
iso_c2d = Cartesian2D([iso_pitch], [iso_pitch])
iso_c2d.set_tiles([isolated_pin])
iso_moc = MOCDriver(iso_c2d)
for i in range(iso_moc.nfsr):
  i_xs = iso_moc.xs(i)
  if i_xs.name == "Fuel":
    iso_moc.set_extern_src(i, 0, 0.)
  elif i_xs.name == "Clad":
    iso_moc.set_extern_src(i, 0, Zirc.potential_xs)
  else:
    iso_moc.set_extern_src(i, 0, Water.potential_xs)
iso_moc.sim_mode = SimulationMode.FixedSource
iso_moc.generate_tracks(128, 0.02, YamamotoTabuchi6())
iso_moc.flux_tolerance = 1.E-6
iso_moc.solve()

# Simulate real geomemtry
radii = [fuel_rad, clad_rad]
mats =  [Fuel, Clad, Mod]
P = SimplePinCell(radii, mats, pitch, pitch)
c2d = Cartesian2D([pitch], [pitch])
c2d.set_tiles([P])
moc = MOCDriver(c2d)
# Set the source
for i in range(moc.nfsr):
  i_xs = moc.xs(i)
  if i_xs.name == "Fuel":
    moc.set_extern_src(i, 0, 0.)
  elif i_xs.name == "Clad":
    moc.set_extern_src(i, 0, Zirc.potential_xs)
  else:
    moc.set_extern_src(i, 0, Water.potential_xs)
moc.generate_tracks(128, 0.02, YamamotoTabuchi6())
moc.flux_tolerance = 1.E-6
moc.sim_mode = SimulationMode.FixedSource
moc.solve()

# Calculate Dancoff correction factor
iso_flux = iso_moc.flux(Vector(0.,0.), Direction(1.,0.), 0)
flux = moc.flux(Vector(0.,0.), Direction(1.,0.), 0)
RR = flux * 1.E5 # Reaction rate in fuel
C = (iso_flux - flux) / iso_flux
print("Dancoff Correction: {}".format(C))
print("Dancoff Factor    : {}".format(1.-C))

# Create fuel xs
Ee = 1. / (2. * fuel_rad)
Fuel = UO2.carlvik_xs(C, Ee, ndl)

# Create cladding xs using a reference dilution
Clad = Zirc.dilution_xs(Zirc.size*[300.], ndl)

# Create moderator xs using a somewhat random dilution (shouldn't matter except a little bit on O16)
Mod = Water.dilution_xs(Water.size*[10.], ndl)

# Create and run full real problem with MOC
radii = [0.5*fuel_rad, 0.8*fuel_rad, fuel_rad, clad_rad, 1.2*clad_rad]
mats  = [Fuel,         Fuel,         Fuel,     Clad,     Mod,         Mod]
P = PinCell(radii, mats, pitch, pitch)
c2d = Cartesian2D([pitch], [pitch])
c2d.set_tiles([P])
moc = MOCDriver(c2d)
moc.generate_tracks(128, 0.005, YamamotoTabuchi6())
moc.flux_tolerance = 1.E-6
moc.solve()

print()
print()

# Solve same problem with CP
radii = [0.5*fuel_rad, 0.8*fuel_rad, fuel_rad, clad_rad, 1.2*clad_rad, pitch/np.sqrt(np.pi)]
mats  = [Fuel,         Fuel,         Fuel,     Clad,     Mod,         Mod]
cell = CylindricalCell(radii, mats)
cell.solve()
cell_flux = CylindricalFluxSolver(cell)
cell_flux.flux_tolerance = 1.E-6
cell_flux.solve()


import pytest
import pytest_check as check
from scarabee import CMFD, CMFDSurfaceCrossing, CMFDSurfaceCrossingType, Direction, Vector
"""
Unit tests for CMFD::tally_current() method

The test mesh consists of a 3x3 rectangular grid with dx, dy = 1.0,1.0

Tests for interior sides:
test_interior_xp
test_interior_xn
test_interior_yn
test_interior_yp

Tests for interior corners:
test_interior_I
test_interior_IV
test_interior_III
test_interior_II

Tests for exterior sides:
test_exterior_xp
test_exterior_xn
test_exterior_yp
test_exterior_yn

Tests for exterior corners:
test_exterior_I_TrueCorner
test_exterior_I_XPCorner
test_exterior_I_YPCorner
test_exterior_IV_TrueCorner
test_exterior_IV_XPCorner
test_exterior_IV_YNCorner
test_exterior_III_TrueCorner
test_exterior_III_XNCorner
test_exterior_III_YNCorner
test_exterior_II_TrueCorner
test_exterior_II_XNCorner
test_exterior_II_YPCorner
"""
class TestTallyCurrent:
    @pytest.fixture
    
    def make_cmfd(self):
        """
        Makes the test grid. Cell indexes are:
        [6, 7, 8]
        [3, 4, 5]
        [0, 1, 2]
        """
        cmfd = CMFD([1.0,1.0,1.0],[1.0,1.0,1.0],[(0,1)])
        return cmfd
    
    def test_interior_xp(self, make_cmfd):

        cmfd = make_cmfd
        aflx = 1.0
        tol = 1e-12
        u = Direction(1.0,0.0)
        r = Vector(0.5-tol,0.0)
        G = 0

        surf = cmfd.get_surface(r,u)
        cmfd.tally_current(aflx,u,G,surf)

        test_current = cmfd.current(G,6)

        assert test_current == 1.0

    def test_interior_yp(self, make_cmfd):

        cmfd = make_cmfd
        aflx = 1.0
        tol = 1e-12
        u = Direction(0.0,1.0)
        r = Vector(0.0,0.5-tol)
        G = 0

        surf = cmfd.get_surface(r,u)
        cmfd.tally_current(aflx,u,G,surf)

        test_current = cmfd.current(G,18)

        assert test_current == 1.0

    def test_interior_xn(self, make_cmfd):

        cmfd = make_cmfd
        aflx = 1.0
        tol = 1e-12
        u = Direction(-1.0,0.0)
        r = Vector(-0.5+tol,0.0)
        G = 0

        surf = cmfd.get_surface(r,u)
        cmfd.tally_current(aflx,u,G,surf)

        test_current = cmfd.current(G,5)

        assert test_current == -1.0
    
    def test_interior_yn(self, make_cmfd):

        cmfd = make_cmfd
        aflx = 1.0
        tol = 1e-12
        u = Direction(0.0,-1.0)
        r = Vector(0.0,-0.5+tol)
        G = 0

        surf = cmfd.get_surface(r,u)
        cmfd.tally_current(aflx,u,G,surf)

        test_current = cmfd.current(G,17)

        assert test_current == -1.0

    def test_interior_I(self,make_cmfd):

        cmfd = make_cmfd
        aflx = 1.0
        tol = 1e-12
        u = Direction(0.707,0.707)
        r = Vector(0.5-tol,0.5-tol)
        G = 0

        surf = cmfd.get_surface(r,u)
        cmfd.tally_current(aflx,u,G,surf)

        tc_1 = cmfd.current(G,6)
        tc_2 = cmfd.current(G,18)
        tc_3 = cmfd.current(G,10)
        tc_4 = cmfd.current(G,22)

        check.equal(tc_1, 0.5)
        check.equal(tc_2, 0.5)
        check.equal(tc_3, 0.5)
        check.equal(tc_4, 0.5)

    def test_interior_IV(self,make_cmfd):

        cmfd = make_cmfd
        aflx = 1.0
        tol = 1e-12
        u = Direction(0.707,-0.707)
        r = Vector(0.5-tol,-0.5+tol)
        G = 0

        surf = cmfd.get_surface(r,u)
        cmfd.tally_current(aflx,u,G,surf)

        tc_1 = cmfd.current(G,6)
        tc_2 = cmfd.current(G,17)
        tc_3 = cmfd.current(G,21)
        tc_4 = cmfd.current(G,2)

        check.equal(tc_1, 0.5)
        check.equal(tc_2, -0.5)
        check.equal(tc_3, -0.5)
        check.equal(tc_4, 0.5)

    def test_interior_III(self,make_cmfd):

        cmfd = make_cmfd
        aflx = 1.0
        tol = 1e-12
        u = Direction(-0.707,-0.707)
        r = Vector(-0.5+tol,-0.5+tol)
        G = 0

        surf = cmfd.get_surface(r,u)
        cmfd.tally_current(aflx,u,G,surf)

        tc_1 = cmfd.current(G,5)
        tc_2 = cmfd.current(G,17)
        tc_3 = cmfd.current(G,1)
        tc_4 = cmfd.current(G,13)

        check.equal(tc_1, -0.5)
        check.equal(tc_2, -0.5)
        check.equal(tc_3, -0.5)
        check.equal(tc_4, -0.5)

    def test_interior_II(self,make_cmfd):

        cmfd = make_cmfd
        aflx = 1.0
        tol = 1e-12
        u = Direction(-0.707,0.707)
        r = Vector(-0.5+tol,0.5-tol)
        G = 0

        surf = cmfd.get_surface(r,u)
        cmfd.tally_current(aflx,u,G,surf)

        tc_1 = cmfd.current(G,9)
        tc_2 = cmfd.current(G,18)
        tc_3 = cmfd.current(G,5)
        tc_4 = cmfd.current(G,14)

        check.equal(tc_1, -0.5)
        check.equal(tc_2, 0.5)
        check.equal(tc_3, -0.5)
        check.equal(tc_4, 0.5)
    
    def test_exterior_yp(self,make_cmfd):
        cmfd = make_cmfd
        aflx = 1.0
        tol = 1e-12
        u = Direction(0.0,-1.0)
        r = Vector(0.0,1.5-tol)
        G = 0

        surf = cmfd.get_surface(r,u)
        cmfd.tally_current(aflx,u,G,surf)

        test_current = cmfd.current(G,19)

        check.equal(test_current, -1.0)
        check.equal(surf.is_valid, True)

    def test_exterior_xp(self,make_cmfd):
        cmfd = make_cmfd
        aflx = 1.0
        tol = 1e-12
        u = Direction(-1.0,0.0)
        r = Vector(1.5-tol,0.0)
        G = 0

        surf = cmfd.get_surface(r,u)
        cmfd.tally_current(aflx,u,G,surf)

        test_current = cmfd.current(G,7)

        check.equal(test_current, -1.0)
        check.equal(surf.is_valid, True)
    
    def test_exterior_yn(self,make_cmfd):
        cmfd = make_cmfd
        aflx = 1.0
        tol = 1e-12
        u = Direction(0.0, 1.0)
        r = Vector(0.0, -1.5+tol)
        G = 0

        surf = cmfd.get_surface(r,u)
        cmfd.tally_current(aflx,u,G,surf)

        test_current = cmfd.current(G,16)

        check.equal(test_current, 1.0)
        check.equal(surf.is_valid, True)

    def test_exterior_xn(self,make_cmfd):
        cmfd = make_cmfd
        aflx = 1.0
        tol = 1e-12
        u = Direction(1.0, 0.0)
        r = Vector(-1.5+tol, 0.0)
        G = 0

        surf = cmfd.get_surface(r,u)
        cmfd.tally_current(aflx,u,G,surf)

        test_current = cmfd.current(G,4)

        check.equal(test_current, 1.0)
        check.equal(surf.is_valid, True)

    def test_exterior_I_TrueCorner(self,make_cmfd):

        cmfd = make_cmfd
        aflx = 1.0
        tol = 1e-12
        u = Direction(-0.707,-0.707)
        r = Vector(1.5-tol,1.5-tol)
        G = 0

        surf = cmfd.get_surface(r,u)
        cmfd.tally_current(aflx,u,G,surf)

        tc_1 = cmfd.current(G,11)
        tc_2 = cmfd.current(G,23)

        check.equal(tc_1, -0.5)
        check.equal(tc_2, -0.5)
    
    def test_exterior_I_XPCorner(self,make_cmfd):

        cmfd = make_cmfd
        aflx = 1.0
        tol = 1e-12
        u = Direction(-0.707,-0.707)
        r = Vector(1.5-tol,0.5-tol)
        G = 0

        surf = cmfd.get_surface(r,u)
        cmfd.tally_current(aflx,u,G,surf)

        tc_1 = cmfd.current(G,11)
        tc_2 = cmfd.current(G,22)
        tc_3 = cmfd.current(G,7)

        check.equal(tc_1, -0.5)
        check.equal(tc_2, -0.5)
        check.equal(tc_3, -0.5)
    
    def test_exterior_I_YPCorner(self,make_cmfd):

        cmfd = make_cmfd
        aflx = 1.0
        tol = 1e-12
        u = Direction(-0.707,-0.707)
        r = Vector(0.5-tol,1.5-tol)
        G = 0

        surf = cmfd.get_surface(r,u)
        cmfd.tally_current(aflx,u,G,surf)

        tc_1 = cmfd.current(G,23)
        tc_2 = cmfd.current(G,10)
        tc_3 = cmfd.current(G,19)

        check.equal(tc_1, -0.5)
        check.equal(tc_2, -0.5)
        check.equal(tc_3, -0.5)
    
    def test_exterior_IV_TrueCorner(self,make_cmfd):

        cmfd = make_cmfd
        aflx = 1.0
        tol = 1e-12
        u = Direction(-0.707,0.707)
        r = Vector(1.5-tol,-1.5+tol)
        G = 0

        surf = cmfd.get_surface(r,u)
        cmfd.tally_current(aflx,u,G,surf)

        tc_1 = cmfd.current(G,20)
        tc_2 = cmfd.current(G,3)

        check.equal(tc_1, 0.5)
        check.equal(tc_2, -0.5)
    
    def test_exterior_IV_YNCorner(self,make_cmfd):

        cmfd = make_cmfd
        aflx = 1.0
        tol = 1e-12
        u = Direction(-0.707, 0.707)
        r = Vector(0.5-tol,-1.5+tol)
        G = 0

        surf = cmfd.get_surface(r,u)
        cmfd.tally_current(aflx,u,G,surf)

        tc_1 = cmfd.current(G,2)
        tc_2 = cmfd.current(G,20)
        tc_3 = cmfd.current(G,16)

        check.equal(tc_1, -0.5)
        check.equal(tc_2, 0.5)
        check.equal(tc_3, 0.5)
    
    def test_exterior_IV_XPCorner(self,make_cmfd):

        cmfd = make_cmfd
        aflx = 1.0
        tol = 1e-12
        u = Direction(-0.707, 0.707)
        r = Vector(1.5-tol,-0.5+tol)
        G = 0

        surf = cmfd.get_surface(r,u)
        cmfd.tally_current(aflx,u,G,surf)

        tc_1 = cmfd.current(G,7)
        tc_2 = cmfd.current(G,21)
        tc_3 = cmfd.current(G,3)

        check.equal(tc_1, -0.5)
        check.equal(tc_2, 0.5)
        check.equal(tc_3, -0.5)
    
    def test_exterior_III_TrueCorner(self,make_cmfd):

        cmfd = make_cmfd
        aflx = 1.0
        tol = 1e-12
        u = Direction(0.707,0.707)
        r = Vector(-1.5+tol,-1.5+tol)
        G = 0

        surf = cmfd.get_surface(r,u)
        cmfd.tally_current(aflx,u,G,surf)

        tc_1 = cmfd.current(G,0)
        tc_2 = cmfd.current(G,12)

        check.equal(tc_1, 0.5)
        check.equal(tc_2, 0.5)

    def test_exterior_III_YNCorner(self,make_cmfd):

        cmfd = make_cmfd
        aflx = 1.0
        tol = 1e-12
        u = Direction(0.707,0.707)
        r = Vector(-0.5+tol,-1.5+tol)
        G = 0

        surf = cmfd.get_surface(r,u)
        cmfd.tally_current(aflx,u,G,surf)

        tc_1 = cmfd.current(G,1)
        tc_2 = cmfd.current(G,16)
        tc_3 = cmfd.current(G,12)

        check.equal(tc_1, 0.5)
        check.equal(tc_2, 0.5)
        check.equal(tc_3, 0.5)
    
    def test_exterior_III_XNCorner(self,make_cmfd):

        cmfd = make_cmfd
        aflx = 1.0
        tol = 1e-12
        u = Direction(0.707,0.707)
        r = Vector(-1.5+tol,-0.5+tol)
        G = 0

        surf = cmfd.get_surface(r,u)
        cmfd.tally_current(aflx,u,G,surf)

        tc_1 = cmfd.current(G,4)
        tc_2 = cmfd.current(G,13)
        tc_3 = cmfd.current(G,0)

        check.equal(tc_1, 0.5)
        check.equal(tc_2, 0.5)
        check.equal(tc_3, 0.5)
    
    def test_exterior_II_TrueCorner(self,make_cmfd):

        cmfd = make_cmfd
        aflx = 1.0
        tol = 1e-12
        u = Direction(0.707,-0.707)
        r = Vector(-1.5+tol,1.5-tol)
        G = 0

        surf = cmfd.get_surface(r,u)
        cmfd.tally_current(aflx,u,G,surf)

        tc_1 = cmfd.current(G,8)
        tc_2 = cmfd.current(G,15)

        check.equal(tc_1, 0.5)
        check.equal(tc_2, -0.5)
    
    def test_exterior_II_YPCorner(self,make_cmfd):

        cmfd = make_cmfd
        aflx = 1.0
        tol = 1e-12
        u = Direction(0.707,-0.707)
        r = Vector(-0.5+tol,1.5-tol)
        G = 0

        surf = cmfd.get_surface(r,u)
        cmfd.tally_current(aflx,u,G,surf)

        tc_1 = cmfd.current(G,19)
        tc_2 = cmfd.current(G,9)
        tc_3 = cmfd.current(G,15)

        check.equal(tc_1, -0.5)
        check.equal(tc_2, 0.5)
        check.equal(tc_3, -0.5)
    
    def test_exterior_II_XNCorner(self,make_cmfd):

        cmfd = make_cmfd
        aflx = 1.0
        tol = 1e-12
        u = Direction(0.707,-0.707)
        r = Vector(-1.5+tol,0.5-tol)
        G = 0

        surf = cmfd.get_surface(r,u)
        cmfd.tally_current(aflx,u,G,surf)

        tc_1 = cmfd.current(G,14)
        tc_2 = cmfd.current(G,4)
        tc_3 = cmfd.current(G,8)

        check.equal(tc_1, -0.5)
        check.equal(tc_2, 0.5)
        check.equal(tc_3, 0.5)

    



from scarabee import FluxCalculator, scarabee_log, LogLevel
import pyPapillonNDL as pndl
import numpy as np


class DilutionTable:
    def __init__(self, dilutions, xs):
        self.dilutions = dilutions
        self.xs = xs  # First index dilution, second is group from high to low energy

        # Ensure that the dilutions are sorted
        if np.all(self.dilutions[:-1] >= self.dilutions[1:]):
            raise RuntimeError("Dilutions are not sorted")

        if self.xs.shape[0] != self.dilutions.shape[0]:
            raise RuntimeError("Number of dilutions and shape of xs array disagree.")

    @property
    def ngroups(self):
        return self.xs.shape[1]

    def group_is_resonant(self, g):
        xs_mid = 0.5 * (
            self.xs[0, g] + self.xs[-1, g]
        )  # Should be the middle value of the xs
        std = np.abs(
            self.xs[0, g] - self.xs[-1, g]
        )  # Should be the absolute difference between min and max dilution

        rel_err = std / xs_mid

        if rel_err <= 0.02:  # 2% threshold for being resonant
            return False

        return True

    def interpolate_xs(self, g, sig_0):
        if g < 0 or g >= self.ngroups:
            raise IndexError("Energy group index out of range.")

        # First, find the LOWER dilution index
        if sig_0 <= self.dilutions[0]:
            return self.xs[0, g]
        elif sig_0 >= self.dilutions[-1]:
            return self.xs[-1, g]

        id = np.searchsorted(self.dilutions, sig_0, side="right") - 1

        sig_0_low = self.dilutions[id]
        sig_0_hi = self.dilutions[id + 1]

        xs_g_low = self.xs[id, g]
        xs_g_hi = self.xs[id + 1, g]

        return ((xs_g_hi - xs_g_low) / (sig_0_hi - sig_0_low)) * (
            sig_0 - sig_0_low
        ) + xs_g_low

    def find_dilution(self, g, xs):
        if g < 0 or g >= self.ngroups:
            raise IndexError("Energy group index is out of range.")

        # If XS is increasing with dilution
        if self.xs[0, g] <= self.xs[-1, g]:
            if xs < self.xs[0, g]:
                raise RuntimeError(
                    "Provided xs is smaller than smallest tabulated dilution."
                )
            elif xs > self.xs[-1, g]:
                raise RuntimeError(
                    "Provided xs is larger than largest tabulated dilution."
                )

            id = np.searchsorted(self.xs[:, g], xs) - 1

        else:
            # XS decreases with dilution
            if xs > self.xs[0, g]:
                raise RuntimeError(
                    "Provided xs is larger than largest tabulated dilution."
                )
            elif xs < self.xs[-1, g]:
                raise RuntimeError(
                    "Provided xs is smaller than smallest tabulated dilution."
                )

            id = np.searchsorted(self.xs[:, g], xs) - 1
            for id in range(self.xs.shape[0] - 1):
                if self.xs[id, g] > xs and xs > self.xs[id + 1, g]:
                    break

        sig_0_low = self.dilutions[id]
        sig_0_hi = self.dilutions[id + 1]

        xs_g_low = self.xs[id, g]
        xs_g_hi = self.xs[id + 1, g]

        return ((sig_0_hi - sig_0_low) / (xs_g_hi - xs_g_low)) * (
            xs - xs_g_low
        ) + sig_0_low


class IRLambdaCalculator:
    # Values taken from ENDF/B-8.0
    AWR_H1 = 0.9991673
    SIG_POT_H1 = 20.436089874720498

    def __init__(
        self,
        Nr,
        awr,
        sig_pot,
        reference_dilution,
        mg_energy_boundaries,
        dilution_table=None,
    ):
        self.Nr = Nr
        self.mg_energy_boundaries = mg_energy_boundaries
        self.dilution_table = dilution_table
        self.reference_dilution = reference_dilution
        self.ir_lambda = None

        # Parameters of nuclide for which we want to generate IR parameters
        self.awr = awr
        self.sig_pot = sig_pot

        # Get parameters for the slowing down calculations
        self.energy_bounds = np.geomspace(1.0e-5, 20.0e6, 500000)
        self.sig_t = np.zeros(self.energy_bounds.size - 1)
        self.sig_s = np.zeros(self.energy_bounds.size - 1)
        for g in range(self.sig_t.size):
            E = 0.5 * (self.energy_bounds[g + 1] + self.energy_bounds[g]) * 1.0e-6
            self.sig_t[g] = self.Nr.total_xs()(E)
            self.sig_s[g] = self.Nr.elastic_xs()(E)

        if self.dilution_table is None:
            self.generate_dilution_table()

    def compute_H1_capture_xs(self, sig_d):
        # Do slowing down calculation
        calc = FluxCalculator(self.energy_bounds, self.sig_t, self.sig_s, self.Nr.awr())
        calc.add_background_nuclide(sig_d, self.AWR_H1)
        calc.solve()

        xs_102 = self.Nr.reaction(102).xs()

        # Setup functions and arrays
        flux = pndl.Tabulated1D(pndl.Interpolation.LinLin, calc.avg_energy, calc.flux)
        sig_mt102_flx = np.array(xs_102.xs())
        enrgy = np.array(xs_102.energy()) * 1.0e6  # Convert energies from MeV to eV
        for i in range(len(enrgy)):
            sig_mt102_flx[i] *= flux(enrgy[i])
        xs_flx = pndl.Tabulated1D(pndl.Interpolation.LinLin, enrgy, sig_mt102_flx)

        # Do integrations for the xs
        mg_sig = np.zeros(self.mg_energy_boundaries.size - 1)
        for g in range(mg_sig.size):
            Ehi = self.mg_energy_boundaries[g]
            Elow = self.mg_energy_boundaries[g + 1]
            mg_sig[g] = xs_flx.integrate(Elow, Ehi) / flux.integrate(Elow, Ehi)

        return mg_sig

    def generate_dilution_table(self):
        scarabee_log(LogLevel.Info, "Generating resonant nuclide dilution table...")
        dilutions = np.geomspace(0.1, 10000.0, 100)
        xs = np.zeros((dilutions.size, self.mg_energy_boundaries.size - 1))

        for i in range(dilutions.size):
            xs[i, :] = self.compute_H1_capture_xs(dilutions[i])

        self.dilution_table = DilutionTable(dilutions, xs)
        scarabee_log(LogLevel.Info, "Resonant nuclide dilution table generated.")

    def solve(self):
        # This ratio is N_H1 / N_r
        R_H1_r = self.reference_dilution / self.SIG_POT_H1

        # We now reduce the ratio of H1 and move it to the nuclide in question
        R_H1 = 0.95 * R_H1_r
        R_nuc = 0.05 * R_H1_r

        # Do slowing down calculation
        calc = FluxCalculator(self.energy_bounds, self.sig_t, self.sig_s, self.Nr.awr())
        calc.add_background_nuclide(R_H1 * self.SIG_POT_H1, self.AWR_H1)
        calc.add_background_nuclide(R_nuc * self.sig_pot, self.awr)
        calc.solve()

        # From flux, compute the MGXS
        xs_102 = self.Nr.reaction(102).xs()

        # Setup functions and arrays
        flux = pndl.Tabulated1D(pndl.Interpolation.LinLin, calc.avg_energy, calc.flux)
        sig_mt102_flx = np.array(xs_102.xs())
        enrgy = np.array(xs_102.energy()) * 1.0e6  # Convert energies from MeV to eV
        for i in range(len(enrgy)):
            sig_mt102_flx[i] *= flux(enrgy[i])
        xs_flx = pndl.Tabulated1D(pndl.Interpolation.LinLin, enrgy, sig_mt102_flx)

        # Do integrations for the xs
        mg_sig = np.zeros(self.mg_energy_boundaries.size - 1)
        for g in range(mg_sig.size):
            Ehi = self.mg_energy_boundaries[g]
            Elow = self.mg_energy_boundaries[g + 1]
            mg_sig[g] = xs_flx.integrate(Elow, Ehi) / flux.integrate(Elow, Ehi)

        # Compute the IR parameters for each group !
        self.ir_lambda = np.zeros(mg_sig.size)
        for g in range(self.ir_lambda.size):
            if (
                self.dilution_table.group_is_resonant(g)
                and self.mg_energy_boundaries[g + 1] > 5.0
            ):  # Only resonant groups and where energy greater than 5 eV
                try:
                    # Get the equivalent background xs
                    sig_0_equiv = self.dilution_table.find_dilution(g, mg_sig[g])
                    self.ir_lambda[g] = (sig_0_equiv - R_H1 * self.SIG_POT_H1) / (
                        R_nuc * self.sig_pot
                    )

                    if self.ir_lambda[g] < 0.0 or 1.0 < self.ir_lambda[g]:
                        Ehi = self.mg_energy_boundaries[g]
                        Elow = self.mg_energy_boundaries[g + 1]
                        mssg = "Reseting IR parameter in group {:d} with energy interval [{:.5E}, {:.5E}] eV from {:.3f} to 1.".format(
                            g + 1, Elow, Ehi, self.ir_lambda[g]
                        )
                        scarabee_log(LogLevel.Warning, mssg)
                        self.ir_lambda[g] = 1.0
                except:
                    mssg = "Group {:d} sig_a: {:.5E}, lowest tabulated sig_a: {:.5E}, highest tabulated sig_a: {:.5E}".format(
                        g + 1,
                        mg_sig[g],
                        self.dilution_table.xs[0, g],
                        self.dilution_table.xs[-1, g],
                    )
                    scarabee_log(LogLevel.Warning, mssg)
                    scarabee_log(LogLevel.Warning, "    Setting IR parameter to 1.")
                    self.ir_lambda[g] = 1.0
            else:
                # Ehi = self.mg_energy_boundaries[g]
                # Elow = self.mg_energy_boundaries[g+1]
                # mssg = "Cannot obtain lambda for group {:d} with energy interval [{:.5E}, {:.5E}] eV. Setting IR parameter to 1.".format(g+1, Elow, Ehi)
                # scarabee_log(LogLevel.Info, mssg)
                self.ir_lambda[g] = 1.0


def generate_U238_U235_ir_lambda(
    U238: pndl.STNeutron,
    U235: pndl.STNeutron,
    awr: float,
    sig_pot: float,
    mg_energy_bounds: np.ndarray,
    dilution: float = 50.0,
):
    lmbda_U238 = IRLambdaCalculator(
        U238,
        awr=awr,
        sig_pot=sig_pot,
        reference_dilution=dilution,
        mg_energy_boundaries=mg_energy_bounds,
    )
    lmbda_U238.solve()

    lmbda_U235 = IRLambdaCalculator(
        U235,
        awr=awr,
        sig_pot=sig_pot,
        reference_dilution=dilution,
        mg_energy_boundaries=mg_energy_bounds,
    )
    lmbda_U235.solve()

    # Start with U238 IR-lambdas as a base
    out_ir_lambda = lmbda_U238.ir_lambda.copy()

    # Now go through and reset with the U235 if we are at 1.
    for g in range(out_ir_lambda.size):
        if out_ir_lambda[g] == 1.0:
            out_ir_lambda[g] = lmbda_U235.ir_lambda[g]

    return out_ir_lambda



import numpy as np
import matplotlib.pyplot as plt

casmo = np.load("casmo_exposure_keff.npy")
casmo2 = np.load("casmo2_exposure_keff.npy")
scarabee = np.load("exposure_keff.npy")

plt.plot(scarabee[0,:], scarabee[1,:], label="Scarabee Style 1 MWd/kg")
plt.plot(casmo[0,:], casmo[1,:],       label="CASMO    Style 1 MWd/kg")
plt.plot(casmo2[0,:], casmo2[1,:],     label="CASMO    Style 2 MWd/kg")
plt.xlabel("Exposure [MWd/kg]")
plt.ylabel("keff")
plt.legend().set_draggable(True)
plt.tight_layout()
plt.show()



import numpy as np
import matplotlib.pyplot as plt
import pyPapillonNDL as pndl
from scarabee import FluxCalculator

def main():
    # Load the continuous energy data for U238
    #ace = pndl.ACE("C:\\Users\\hunte\\Documents\\nuclear_data\\ace\\endf8.0\\abeille_binary\\data\\neutron_dir\\U\\U238\\U238.293.6.ace.bin", pndl.ACEType.BINARY)
    ace = pndl.ACE("/mnt/c/Users/hunte/Documents/nuclear_data/ace/endf8.0/abeille_binary/data/neutron_dir/U/U238/U238.293.6.ace.bin", pndl.ACEType.BINARY)
    Nuc = pndl.STNeutron(ace)

    # Make an array for the ultra-fine slowing down calculation
    energy_bounds = np.geomspace(1.E-5, 20.E6, 500000)
    sig_t = np.zeros(energy_bounds.size-1)
    sig_s = np.zeros(energy_bounds.size-1)
    for g in range(sig_t.size):
        E = 0.5*(energy_bounds[g+1] + energy_bounds[g]) * 1.E-6
        sig_t[g] = Nuc.total_xs()(E)
        sig_s[g] = Nuc.elastic_xs()(E)
    
    # Atomic Weight Ratio of H1
    AWR_H1 = 0.9991673

    # Perform slowing down calculation for U238 with a 50 barn background xs from H1
    calc = FluxCalculator(energy_bounds, sig_t, sig_s, Nuc.awr())
    calc.add_background_nuclide(50., AWR_H1)
    calc.solve()

    plt.plot(calc.avg_energy, calc.flux)
    plt.xscale('log')
    plt.yscale('log')
    plt.xlabel('Energy [eV]')
    plt.ylabel('Flux [Arb. Units]')
    plt.tight_layout()
    plt.show()
    
if __name__ == '__main__':
    main()



# This is problem 2 from the following paper:
# N. SUGIMURA and A. YAMAMOTO, “Evaluation of Dancoff Factors in Complicated
# Geometry using the Method of Characteristics,” J. Nucl. Sci. Technol.,
# vol. 43, no. 10, pp. 1182–1187, 2006, doi: 10.1080/18811248.2006.9711210.

from scarabee import *
import numpy as np
np.set_printoptions(linewidth=200)

Et = np.array([1.E5])
Ea = np.array([1.E5])
Es = np.array([[0.]])
Fuel = CrossSection(Et, Ea, Es, "Fuel")

Mod_pot_xs = 1.05
Et = np.array([Mod_pot_xs])
Ea = np.array([Mod_pot_xs])
Mod = CrossSection(Et, Ea, Es, "Mod")

Clad_pot_xs = 0.30
Et = np.array([Clad_pot_xs])
Ea = np.array([Clad_pot_xs])
Clad = CrossSection(Et, Ea, Es, "Clad")

Thmb_pot_xs = 0.30
Et = np.array([Thmb_pot_xs])
Ea = np.array([Thmb_pot_xs])
Thmb = CrossSection(Et, Ea, Es, "Thmb")

# Define Cells
pitch = 1.26
iso_pitch = 7.*pitch
fuel_rad = 0.41
clad_rad = 0.48
mod_rad = 0.57
thmb_rad = 0.61

# Simulate isolated geometry
radii = [fuel_rad, clad_rad]
mats =  [Fuel, Clad, Mod]
isolated_pin = SimplePinCell(radii, mats, iso_pitch, iso_pitch)
iso_c2d = Cartesian2D([iso_pitch], [iso_pitch])
iso_c2d.set_tiles([isolated_pin])
iso_moc = MOCDriver(iso_c2d)
for i in range(iso_moc.nfsr):
  i_xs = iso_moc.xs(i)
  if i_xs.name == "Fuel":
    iso_moc.set_extern_src(i, 0, 0.)
  elif i_xs.name == "Clad":
    iso_moc.set_extern_src(i, 0, Clad_pot_xs)
  else:
    iso_moc.set_extern_src(i, 0, Mod_pot_xs)
iso_moc.sim_mode = SimulationMode.FixedSource
iso_moc.generate_tracks(64, 0.01, Legendre12())
iso_moc.flux_tolerance = 1.E-6
iso_moc.solve()

# Simulate real geomemtry
radii = [fuel_rad, clad_rad]
mats =  [Fuel, Clad, Mod]
F = SimplePinCell(radii, mats, pitch, pitch)
radii = [mod_rad, thmb_rad]
mats = [Mod, Thmb, Mod]
G = SimplePinCell(radii, mats, pitch, pitch)
c2d = Cartesian2D(17*[pitch], 17*[pitch])
c2d.set_tiles([F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,
               F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,
               F,F,F,F,F,G,F,F,G,F,F,G,F,F,F,F,F,
               F,F,F,G,F,F,F,F,F,F,F,F,F,G,F,F,F,
               F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,
               F,F,G,F,F,G,F,F,G,F,F,G,F,F,G,F,F,
               F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,
               F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,
               F,F,G,F,F,G,F,F,G,F,F,G,F,F,G,F,F,
               F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,
               F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,
               F,F,G,F,F,G,F,F,G,F,F,G,F,F,G,F,F,
               F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,
               F,F,F,G,F,F,F,F,F,F,F,F,F,G,F,F,F,
               F,F,F,F,F,G,F,F,G,F,F,G,F,F,F,F,F,
               F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,
               F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F])
moc = MOCDriver(c2d)
# Set the source
for i in range(moc.nfsr):
  i_xs = moc.xs(i)
  if i_xs.name == "Fuel":
    moc.set_extern_src(i, 0, 0.)
  elif i_xs.name == "Clad":
    moc.set_extern_src(i, 0, Clad_pot_xs)
  elif i_xs.name == "Thmb":
    moc.set_extern_src(i, 0, Thmb_pot_xs)
  else:
    moc.set_extern_src(i, 0, Mod_pot_xs)
moc.generate_tracks(64, 0.01, Legendre12())
moc.flux_tolerance = 1.E-6
moc.sim_mode = SimulationMode.FixedSource
moc.solve()

# Calculate Dancoff correction factor
iso_flux = iso_moc.flux(Vector(0.,0.), Direction(1.,0.), 0)

D = np.zeros((17,17))
for j in range(17):
  y = moc.y_max - (j+0.5)*pitch
  for i in range(17):
    x = moc.x_min + (i+0.5)*pitch
    xs = moc.xs(Vector(x,y), Direction(1.,0.))
    if xs.name == "Fuel":
      flux = moc.flux(Vector(x,y), Direction(1.,0.), 0)
      C = (iso_flux - flux) / iso_flux
      D[j,i] = 1. - C

print()
for i in range(9):
  print(D[i+8, 8:i+9])


from ._scarabee import *
from . import _scarabee

__author__ = _scarabee.__author__
__copyright__ = _scarabee.__copyright__
__license__ = _scarabee.__license__
__maintainer__ = _scarabee.__maintainer__
__email__ = _scarabee.__email__
__version__ = _scarabee.__version__



import pytest
import pytest_timeout
import numpy as np
from scarabee import *
"""
Integration test for CMFD Fixed Source mode on a point source in a water assembly
"""

class TestCMFDWaterFS:

    # This problem should never take this long to run
    # if it takes longer, most likely didn't converge
    @pytest.mark.timeout(200)
    def test_water_fs(self):
        '''
        Test fixed point source problem with CMFD, isotropic scattering
        '''
        Et = np.array([1.59206E-01, 4.12970E-01, 5.90310E-01, 5.84350E-01, 7.18000E-01, 1.25445E+00, 2.65038E+00])
        Ea = np.array([6.01050E-04, 1.57930E-05, 3.37160E-04, 1.94060E-03, 5.74160E-03, 1.50010E-02, 3.72390E-02])
        Es = np.array([[4.44777E-02, 1.13400E-01, 7.23470E-04, 3.74990E-06, 5.31840E-08, 0.00000E+00, 0.00000E+00],
                    [0.00000E+00, 2.82334E-01, 1.29940E-01, 6.23400E-04, 4.80020E-05, 7.44860E-06, 1.04550E-06],
                    [0.00000E+00, 0.00000E+00, 3.45256E-01, 2.24570E-01, 1.69990E-02, 2.64430E-03, 5.03440E-04],
                    [0.00000E+00, 0.00000E+00, 0.00000E+00, 9.10284E-02, 4.15510E-01, 6.37320E-02, 1.21390E-02],
                    [0.00000E+00, 0.00000E+00, 0.00000E+00, 7.14370E-05, 1.39138E-01, 5.11820E-01, 6.12290E-02],
                    [0.00000E+00, 0.00000E+00, 0.00000E+00, 0.00000E+00, 2.21570E-03, 6.99913E-01, 5.37320E-01],
                    [0.00000E+00, 0.00000E+00, 0.00000E+00, 0.00000E+00, 0.00000E+00, 1.32440E-01, 2.48070E+00]])
        H2Oxs = CrossSection(Et, Ea, Es)

        # Define Cells
        pitch = 1.26

        NW = 6 # Number of empty water cells in a pin cell
        dx = [pitch / NW] * NW
        WC = EmptyCell(H2Oxs, pitch / NW, pitch / NW)
        WT = Cartesian2D(dx, dx)
        WT.set_tiles([WC]*NW*NW)

        # Water assembly
        dx = [pitch]*17
        WAS = Cartesian2D(dx, dx)
        WAS.set_tiles([WT]*17*17)

        # To use pincell level cells for this problem larsen correction must be enabled
        dxc = [pitch]*17

        moc = MOCDriver(WAS)
        moc.x_min_bc = BoundaryCondition.Vacuum
        moc.x_max_bc = BoundaryCondition.Vacuum
        moc.y_min_bc = BoundaryCondition.Vacuum
        moc.y_max_bc = BoundaryCondition.Vacuum
        moc.cmfd = CMFD(dxc,dxc,[[0,0], [1,1], [2,2], [3,3], [4,4], [5,5], [6,6]])
        # Works with several damping factors, but may become unstable
        # with larger values
        moc.cmfd.damping = 0.5
        moc.cmfd.flux_limiting = True
        # Optimal number of MOC iterations to skip for this problem
        moc.cmfd.skip_moc_iterations = 14
        moc.generate_tracks(32, 0.05, YamamotoTabuchi6())
        moc.set_extern_src(Vector(0.,0.), Direction(0.,1.), 0, 1.)
        moc.flux_tolerance = 1.E-5
        moc.sim_mode = SimulationMode.FixedSource
        moc.solve()

        # Test flux in each group at a single FSR
        test_point = Vector(0.1,0.1)
        test_dir = Direction(0.,1.)
        assert moc.flux(test_point, test_dir, 0) == pytest.approx(0.661793, 1E-4)
        assert moc.flux(test_point, test_dir, 1) == pytest.approx(0.086848, 1E-4)
        assert moc.flux(test_point, test_dir, 2) == pytest.approx(0.022218, 1E-4)
        assert moc.flux(test_point, test_dir, 3) == pytest.approx(0.007779, 1E-4)
        assert moc.flux(test_point, test_dir, 4) == pytest.approx(0.005393, 1E-4)
        assert moc.flux(test_point, test_dir, 5) == pytest.approx(0.017878, 1E-4)
        assert moc.flux(test_point, test_dir, 6) == pytest.approx(0.054445, 1E-4)



from scarabee import (
    NDLibrary,
    MaterialComposition,
    Material,
    Fraction,
    DensityUnits,
    set_output_file,
)
from scarabee.reseau import FuelPin, GuideTube, BurnablePoisonRod, PWRAssembly, Symmetry

name = "F31_20"

set_output_file(name + "_out.txt")

ndl = NDLibrary()

# Define all Materials
Fuel31Comp = MaterialComposition(Fraction.Atoms, name="Fuel 3.1%")
Fuel31Comp.add_leu(3.1, 1.0)
Fuel31Comp.add_element("O", 2.0)
Fuel31 = Material(Fuel31Comp, 575.0, 10.30166, DensityUnits.g_cm3, ndl)

CladComp = MaterialComposition(Fraction.Weight, name="Zircaloy 4")
CladComp.add_element("O", 0.00125)
CladComp.add_element("Cr", 0.0010)
CladComp.add_element("Fe", 0.0021)
CladComp.add_element("Zr", 0.98115)
CladComp.add_element("Sn", 0.0145)
Clad = Material(CladComp, 575.0, 6.55, DensityUnits.g_cm3, ndl)

HeComp = MaterialComposition(Fraction.Atoms, name="He Gas")
HeComp.add_element("He", 1.0)
He = Material(HeComp, 575.0, 0.0015981, DensityUnits.g_cm3, ndl)

SS304Comp = MaterialComposition(Fraction.Weight, name="SS304")
SS304Comp.add_element("Si", 0.0060)
SS304Comp.add_element("Cr", 0.1900)
SS304Comp.add_element("Mn", 0.0200)
SS304Comp.add_element("Fe", 0.6840)
SS304Comp.add_element("Ni", 0.1000)
SS304 = Material(SS304Comp, 575.0, 8.03, DensityUnits.g_cm3, ndl)

BSiGlassComp = MaterialComposition(Fraction.Weight, name="Borosilicate Glass")
BSiGlassComp.add_element("O", 0.5481)
BSiGlassComp.add_element("Si", 0.3787)
BSiGlassComp.add_element("Al", 0.0344)
BSiGlassComp.add_nuclide("B10", 0.0071)
BSiGlassComp.add_nuclide("B11", 0.0317)
BSiGlass = Material(BSiGlassComp, 575.0, 2.26, DensityUnits.g_cm3, ndl)

AirComp = MaterialComposition(Fraction.Atoms, name="Air")
AirComp.add_element("O", 0.2095)
AirComp.add_element("N", 0.7809)
AirComp.add_element("Ar", 0.00933)
AirComp.add_element("C", 0.00027)
Air = Material(AirComp, 575.0, 0.000616, DensityUnits.g_cm3, ndl)

# Define a guide tube
gt = GuideTube(inner_radius=0.56134, outer_radius=0.60198, clad=Clad)

# Define guide tube with a burnable poison rod
bpr = BurnablePoisonRod(
    center=Air,
    clad=SS304,
    gap=He,
    poison=BSiGlass,
    center_radius=0.214,
    inner_clad_radius=0.23051,
    inner_gap_radius=0.24130,
    poison_radius=0.42672,
    outer_gap_radius=0.43688,
    outer_clad_radius=0.48387,
)

bp = GuideTube(inner_radius=0.56134, outer_radius=0.60198, clad=Clad, fill=bpr)

# Define fuel pin
fp = FuelPin(
    fuel=Fuel31,
    fuel_radius=0.39218,
    gap=He,
    gap_radius=0.40005,
    clad=Clad,
    clad_radius=0.45720,
)

cells = [
    [fp, fp, fp, fp, fp, fp, fp, fp, fp],
    [fp, fp, fp, fp, fp, fp, fp, fp, fp],
    [bp, fp, fp, bp, fp, fp, fp, fp, fp],
    [fp, fp, fp, fp, fp, bp, fp, fp, fp],
    [fp, fp, fp, fp, fp, fp, fp, fp, fp],
    [gt, fp, fp, bp, fp, fp, bp, fp, fp],
    [fp, fp, fp, fp, fp, fp, fp, fp, fp],
    [fp, fp, fp, fp, fp, fp, fp, fp, fp],
    [gt, fp, fp, gt, fp, fp, bp, fp, fp],
]

# Define assembly
asmbly = PWRAssembly(
    pitch=1.25984,
    assembly_pitch=21.50364,
    shape=(17, 17),
    symmetry=Symmetry.Quarter,
    moderator_pressure=15.5132,
    moderator_temp=575.0,
    boron_ppm=975.0,
    cells=cells,
    ndl=ndl,
)

asmbly.solve()
asmbly.diffusion_data.save(name + ".bin")



from .._scarabee import DiffusionCrossSection
import numpy as np
from typing import Optional, Union


class NodalFlux1D:
    def __init__(
        self,
        x_min: float,
        x_max: float,
        keff: float,
        xs: DiffusionCrossSection,
        avg_flx: np.ndarray,
        j_neg: np.ndarray,
        j_pos: np.ndarray,
    ):
        """
        Performs the 1D nodal diffusion calculation using the Nodal Expansion
        Method (NEM) with reference currents and average fluxes. This
        represents the transverse integrated homogeneous nodal flux solution.
        The primary use of this class is for computing the homogeneous flux
        that is required to define discontinuity factors for generalized
        equivalence theory.

        Paramters
        ---------
        x_min : float
             Position of the negative surface of the node.
        x_max : float
             Position of the positive surface of the node.
        keff : float
             Multiplication factor from reference Sn calculation.
        xs : DiffusionCrossSection
             Diffusion cross sections homogenized for the node.
        avg_flx : np.ndarray
             Average flux in each group within the node from reference Sn
             calculation.
        j_neg : np.ndarray
             Reference net current in each group on the negative boundary
             from the reference calculation.
        j_pos : np.ndarray
             Reference net current in each group on the positive boundary
             from the reference calculation.
        """
        if x_min >= x_max:
            raise ValueError("The value of x_min must be < x_max.")
        self.x_min = x_min
        self.x_max = x_max
        dx = x_max - x_min  # Only needed locally to compute coefficients

        if keff <= 0.0 or 2.0 <= keff:
            raise ValueError("The value of keff must be in the interval (0, 2).")

        self.ngroups = xs.ngroups

        # Checks for all arrays
        if avg_flx.ndim != 1:
            raise ValueError("The array avg_flux must be 1D.")
        if j_neg.ndim != 1:
            raise ValueError("The array j_neg must be 1D.")
        if j_pos.ndim != 1:
            raise ValueError("The array j_pos must be 1D.")

        if avg_flx.size != xs.ngroups:
            raise ValueError(
                "The length of avg_flux does not agree with number of groups in xs."
            )
        if j_neg.size != xs.ngroups:
            raise ValueError(
                "The length of j_neg does not agree with number of groups in xs."
            )
        if j_pos.size != xs.ngroups:
            raise ValueError(
                "The length of j_neg does not agree with number of groups in xs."
            )

        # Perform the static nodal calculation for all groups
        NG = self.ngroups
        Na = NG * 4  # number of a coefficients to solve for
        invs_dx = 1.0 / dx
        A = np.zeros((Na, Na))  # Matrix to hold all coefficients
        b = np.zeros((Na))  # Results array

        # Load the coefficient matrix and results array
        j = 0  # Matrix row
        for g in range(NG):
            Dg = xs.D(g)
            Dg_dx = Dg * invs_dx
            # Use this for Et instead of 1/(3D), as that gave bad results.
            Et = xs.Ea(g) + xs.Es(g)
            chi_g_keff = xs.chi(g) / keff
            Erf_g = Et - xs.Es(g, g) - chi_g_keff * xs.vEf(g)

            # Each group has 4 equations. See reference [1].

            # Eq 2.54
            A[j, g * 4 + 2] -= 0.5 * Dg_dx * invs_dx
            A[j, g * 4 + 0] += Erf_g / 12.0
            A[j, g * 4 + 2] -= 0.1 * (Erf_g / 12.0)
            for gg in range(NG):
                if gg == g:
                    continue
                A[j, gg * 4 + 0] -= xs.Es(gg, g) / 12.0
                A[j, gg * 4 + 2] += 0.1 * (xs.Es(gg, g) / 12.0)
                A[j, gg * 4 + 0] -= chi_g_keff * (xs.vEf(gg) / 12.0)
                A[j, gg * 4 + 2] += 0.1 * chi_g_keff * (xs.vEf(gg) / 12.0)
            b[j] = 0.0
            j += 1

            # Eq 2.55
            A[j, g * 4 + 3] -= 0.2 * Dg_dx * invs_dx
            A[j, g * 4 + 1] += Erf_g / 20.0
            A[j, g * 4 + 3] -= Erf_g / (20.0 * 35.0)
            for gg in range(NG):
                if gg == g:
                    continue
                A[j, gg * 4 + 1] -= xs.Es(gg, g) / 20.0
                A[j, gg * 4 + 3] += xs.Es(gg, g) / (20.0 * 35.0)
                A[j, gg * 4 + 1] -= chi_g_keff * (xs.vEf(gg) / 20.0)
                A[j, gg * 4 + 3] += chi_g_keff * (xs.vEf(gg) / (20.0 * 35.0))
            b[j] = 0.0
            j += 1

            # Eq 2.57 and 2.58 have an error in [1]. They are both missing a
            # factor of 3 on the a2 term. The correct form can be found from
            # Lawrence in [2].
            # Eq 2.57
            A[j, g * 4 + 0] -= Dg_dx
            A[j, g * 4 + 1] += 3.0 * Dg_dx
            A[j, g * 4 + 2] -= 0.5 * Dg_dx
            A[j, g * 4 + 3] += 0.2 * Dg_dx
            b[j] = j_neg[g]
            j += 1

            # Eq 2.58
            A[j, g * 4 + 0] -= Dg_dx
            A[j, g * 4 + 1] -= 3.0 * Dg_dx
            A[j, g * 4 + 2] -= 0.5 * Dg_dx
            A[j, g * 4 + 3] -= 0.2 * Dg_dx
            b[j] = j_pos[g]
            j += 1

        a_tmp = np.linalg.solve(A, b)
        self.a = np.zeros((NG, 5))
        for g in range(NG):
            self.a[g, 0] = avg_flx[g]
            self.a[g, 1:] = a_tmp[g * 4 : g * 4 + 4]

    def __call__(self, x: float, g: int) -> float:
        if g < 0:
            raise RuntimeError("Group index g must be >= 0.")

        if g >= self.ngroups:
            raise RuntimeError("Group index g is out of range.")

        # Commented out to permit use with an array
        """
        if x < self.x_min:
            raise RuntimeError("x value is less than lower boundary.")
        
        if x > self.x_max:
            raise RuntimeError("x value is greater than upper boundary.")
        """

        u = ((x - self.x_min) / (self.x_max - self.x_min)) - 0.5
        u2 = u * u
        f1 = u
        f2 = 3.0 * u2 - 0.25
        f3 = (u2 - 0.25) * u
        f4 = (u2 - 0.25) * (u2 - 0.05)

        flx = (
            self.a[g, 0]
            + self.a[g, 1] * f1
            + self.a[g, 2] * f2
            + self.a[g, 3] * f3
            + self.a[g, 4] * f4
        )

        return flx

    def pos_surf_flux(self, g: Optional[int] = None) -> Union[np.ndarray, float]:
        if g is None:
            return self.a[:, 0] + 0.5 * self.a[:, 1] + 0.5 * self.a[:, 2]

        if g < 0:
            raise RuntimeError("Group index g must be >= 0.")

        if g >= self.ngroups:
            raise RuntimeError("Group index g is out of range.")

        return self.a[g, 0] + 0.5 * self.a[g, 1] + 0.5 * self.a[g, 2]

    def neg_surf_flux(self, g: Optional[int] = None) -> Union[np.ndarray, float]:
        if g is None:
            return self.a[:, 0] - 0.5 * self.a[:, 1] + 0.5 * self.a[:, 2]

        if g < 0:
            raise RuntimeError("Group index g must be >= 0.")

        if g >= self.ngroups:
            raise RuntimeError("Group index g is out of range.")

        return self.a[g, 0] - 0.5 * self.a[g, 1] + 0.5 * self.a[g, 2]


class NodalFlux2D:
    def __init__(
        self,
        dx: float,
        dy: float,
        keff: float,
        xs: DiffusionCrossSection,
        avg_flx: np.ndarray,
        j_x_neg: np.ndarray,
        j_x_pos: np.ndarray,
        j_y_neg: np.ndarray,
        j_y_pos: np.ndarray,
    ):
        """
        Performs the 2D nodal diffusion calculation using the Nodal Expansion
        Method (NEM) with reference currents and average fluxes. This
        represents the homogeneous nodal flux solution. The primary use of this
        class is for computing the homogeneous flux that is required to define
        corner discontinuity factors for flux reconstruction.

        By default, the coupled x-y terms are neglected, as this requires
        knowledge of the average corner fluxes based on the adjacent nodes. Due
        to this restriction, the coefficients for the coupled x-y terms must be
        computed separately. Examine the source code in
        NEMDriver::fit_node_recon_params_corners for an example of how this
        should be done. It follows the ANOVA-HDMR Decomposition method [3].

        Paramters
        ---------
        dx : float
             Width of the node along the x axis.
        dy : float
             Width of the node along the y axis.
        keff : float
             Multiplication factor from reference Sn calculation.
        xs : DiffusionCrossSection
             Diffusion cross sections homogenized for the node.
        avg_flx : np.ndarray
             Average flux in each group within the node from reference Sn
             calculation.
        j_x_neg : np.ndarray
             Reference net current in each group on the negative x boundary
             from the reference calculation.
        j_x_pos : np.ndarray
             Reference net current in each group on the positive x boundary
             from the reference calculation.
        j_y_neg : np.ndarray
             Reference net current in each group on the negative y boundary
             from the reference calculation.
        j_y_pos : np.ndarray
             Reference net current in each group on the positive y boundary
             from the reference calculation.
        """
        self.keff = keff
        self.ngroups = xs.ngroups
        NG = self.ngroups

        self.phi_0 = avg_flx  # f0
        self.eps = np.zeros(NG)

        # fx
        self.ax0 = np.zeros(NG)
        self.ax1 = np.zeros(NG)
        self.ax2 = np.zeros(NG)
        self.bx1 = np.zeros(NG)
        self.bx2 = np.zeros(NG)

        # fy
        self.ay0 = np.zeros(NG)
        self.ay1 = np.zeros(NG)
        self.ay2 = np.zeros(NG)
        self.by1 = np.zeros(NG)
        self.by2 = np.zeros(NG)

        # fxy
        self.cxy11 = np.zeros(NG)
        self.cxy12 = np.zeros(NG)
        self.cxy21 = np.zeros(NG)
        self.cxy22 = np.zeros(NG)

        self.invs_dx = 1.0 / dx
        self.invs_dy = 1.0 / dy
        self.zeta_x = np.zeros(NG)
        self.zeta_y = np.zeros(NG)

        self.flux_x = NodalFlux1D(
            -0.5 * dx, 0.5 * dx, self.keff, xs, avg_flx, j_x_neg, j_x_pos
        )
        self.flux_y = NodalFlux1D(
            -0.5 * dy, 0.5 * dy, self.keff, xs, avg_flx, j_y_neg, j_y_pos
        )

        self._initialize_params_no_cross_terms(xs, j_x_pos, j_x_neg, j_y_pos, j_y_neg)

    @property
    def dx(self) -> float:
        return 1.0 / self.invs_dx

    @property
    def dy(self) -> float:
        return 1.0 / self.invs_dy

    def _initialize_params_no_cross_terms(
        self, xs: DiffusionCrossSection, Jxp, Jxm, Jyp, Jym
    ) -> None:
        sinhc = lambda x: np.sinh(x) / x

        for g in range(self.ngroups):
            D = xs.D(g)
            Er = xs.Er(g)
            self.eps[g] = np.sqrt(Er / D)

            flx_avg = self.phi_0[g]
            flx_xp = self.flux_x.pos_surf_flux(g)
            flx_xm = self.flux_x.neg_surf_flux(g)
            flx_yp = self.flux_y.pos_surf_flux(g)
            flx_ym = self.flux_y.neg_surf_flux(g)

            # Initial base matrix for finding fx, fy, and fz coefficients
            M = np.array(
                [
                    [0.0, 0.0, 1.0, 1.0],
                    [0.0, 0.0, -1.0, 1.0],
                    [0.0, 0.0, 1.0, 3.0],
                    [0.0, 0.0, 1.0, -3.0],
                ]
            )
            b = np.zeros(4)

            # Determine fx coefficients
            zeta_x = 0.5 * self.eps[g] * self.dx
            M[0, 0] = np.cosh(zeta_x) - sinhc(zeta_x)
            M[0, 1] = np.sinh(zeta_x)
            M[1, 0] = M[0, 0]
            M[1, 1] = -M[0, 1]
            M[2, 0] = zeta_x * np.sinh(zeta_x)
            M[2, 1] = zeta_x * np.cosh(zeta_x)
            M[3, 0] = -M[2, 0]
            M[3, 1] = M[2, 1]
            b[0] = flx_xp - flx_avg
            b[1] = flx_xm - flx_avg
            b[2] = -0.5 * Jxp[g] * self.dx / D
            b[3] = -0.5 * Jxm[g] * self.dx / D
            fu_coeffs = np.linalg.solve(M, b)
            self.ax1[g] = fu_coeffs[0]
            self.ax2[g] = fu_coeffs[1]
            self.bx1[g] = fu_coeffs[2]
            self.bx2[g] = fu_coeffs[3]
            self.ax0[g] = -self.ax1[g] * sinhc(zeta_x)
            self.zeta_x[g] = zeta_x

            # Determine fy coefficients
            zeta_y = 0.5 * self.eps[g] * self.dy
            M[0, 0] = np.cosh(zeta_y) - sinhc(zeta_y)
            M[0, 1] = np.sinh(zeta_y)
            M[1, 0] = M[0, 0]
            M[1, 1] = -M[0, 1]
            M[2, 0] = zeta_y * np.sinh(zeta_y)
            M[2, 1] = zeta_y * np.cosh(zeta_y)
            M[3, 0] = -M[2, 0]
            M[3, 1] = M[2, 1]
            b[0] = flx_yp - flx_avg
            b[1] = flx_ym - flx_avg
            b[2] = -0.5 * Jyp[g] * self.dy / D
            b[3] = -0.5 * Jym[g] * self.dy / D
            fu_coeffs = np.linalg.solve(M, b)
            self.ay1[g] = fu_coeffs[0]
            self.ay2[g] = fu_coeffs[1]
            self.by1[g] = fu_coeffs[2]
            self.by2[g] = fu_coeffs[3]
            self.ay0[g] = -self.ay1[g] * sinhc(zeta_y)
            self.zeta_y[g] = zeta_y

    def __call__(self, x, y) -> np.ndarray:
        return self.phi_0 + self.fx(x) + self.fy(y) + self.fxy(x, y)

    def flux_xy_no_cross(self, x, y) -> np.ndarray:
        return self.phi_0 + self.fx(x) + self.fy(y)

    def fx(self, x) -> np.ndarray:
        return (
            self.ax0
            + self.ax1 * np.cosh(self.eps * x)
            + self.ax2 * np.sinh(self.eps * x)
            + self.bx1 * self.p1(2.0 * x * self.invs_dx)
            + self.bx2 * self.p2(2.0 * x * self.invs_dx)
        )

    def fy(self, y) -> np.ndarray:
        return (
            self.ay0
            + self.ay1 * np.cosh(self.eps * y)
            + self.ay2 * np.sinh(self.eps * y)
            + self.by1 * self.p1(2.0 * y * self.invs_dy)
            + self.by2 * self.p2(2.0 * y * self.invs_dy)
        )

    def fxy(self, x, y) -> np.ndarray:
        x *= 2.0 * self.invs_dx
        y *= 2.0 * self.invs_dy
        p1x = self.p1(x)
        p2x = self.p2(x)
        p1y = self.p1(y)
        p2y = self.p2(y)
        return (
            self.cxy11 * p1x * p1y
            + self.cxy12 * p1x * p2y
            + self.cxy21 * p2x * p1y
            + self.cxy22 * p2x * p2y
        )

    def p1(self, xi) -> float:
        return xi

    def p2(self, xi) -> float:
        return 0.5 * (3.0 * xi * xi - 1.0)


# REFERENCES
#
# [1] S. Machach, “Étude des techniques d’équivalence nodale appliquées aux
#     modèles de réflecteurs dans les réacteurs à eau pressurisée,”
#     Polytechnique Montréal, 2022.
#
# [2] R. D. Lawrence, “Progress in nodal methods for the solution of the
#     neutron diffusion and transport equations,” Prog. Nucl. Energ., vol. 17,
#     no. 3, pp. 271–301, 1986, doi: 10.1016/0149-1970(86)90034-x.
#
# [3] P. M. Bokov, D. Botes, R. H. Prinsloo, and D. I. Tomašević, “A
#     Multigroup Homogeneous Flux Reconstruction Method Based on the
#     ANOVA-HDMR Decomposition,” Nucl. Sci. Eng., vol. 197, no. 2,
#     pp. 308–332, 2023, doi: 10.1080/00295639.2022.2108654.



from .._scarabee import (
    NDLibrary,
    Material,
    CrossSection,
    PinCellType,
    SimplePinCell,
    PinCell,
    MOCDriver,
    DepletionChain,
)
from .burnable_poison_rod import BurnablePoisonRod
import numpy as np
from typing import Optional, List
import copy


class GuideTube:
    """
    Represents an empty guide tube for a PWR.

    Parameters
    ----------
    clad : Material
        Material which describes the cladding composition, density, and
        temperature.
    inner_radius : float
        Inner radius of the guide tube.
    outer_radius : float
        Outer radius of the guide tube.
    fill : BurnablePoisonRod, optional
        Optional burnable poison rod which can be placed inside the guide tube.
        Default value is None.

    Attributes
    ----------
    clad : Material
        Material which describes the cladding composition, density, and
        temperature.
    inner_radius : float
        Inner radius of the guide tube.
    outer_radius : float
        Outer radius of the guide tube.
    fill : BurnablePoisonRod, optional
        Optional burnable poison rod which can be placed inside the guide tube.
    clad_dancoff_corrections : list of float
        Dancoff corrections to be used when self-shielding the cladding at each
        depletion time step.
    empty : bool
        True if fill is None and False otherwise.
    """

    def __init__(
        self,
        clad: Material,
        inner_radius: float,
        outer_radius: float,
        fill: Optional[BurnablePoisonRod] = None,
    ):
        if inner_radius >= outer_radius:
            raise ValueError("Inner radius must be > outer radius.")

        self._clad = copy.deepcopy(clad)
        self._inner_radius = inner_radius
        self._outer_radius = outer_radius
        self._fill = copy.deepcopy(fill)

        # Make sure the control rod of poison rod would actually fit inside the
        # guide tube. Do this by checking the radii.
        if self.fill is not None:
            if isinstance(self.fill, BurnablePoisonRod):
                if self.fill.outer_clad_radius >= self.inner_radius:
                    raise ValueError(
                        "The burnable poison rod is too large for the guide tube."
                    )
            else:
                raise TypeError("Unknown fill object placed in guide tube.")

        # ======================================================================
        # DANCOFF CORRECTION CALCULATION DATA
        # ----------------------------------------------------------------------

        # Initialize empty list of Dancoff corrections for the cladding
        self._clad_dancoff_corrections: List[float] = []

        # Initialize empty variables for Dancoff correction calculations.
        # These are all kept private.
        self._clad_dancoff_xs: CrossSection = CrossSection(
            np.array([self.clad.potential_xs]),
            np.array([self.clad.potential_xs]),
            np.array([[0.0]]),
            "Clad",
        )

        self._clad_isolated_dancoff_fsr_ids = []
        self._clad_isolated_dancoff_fsr_inds = []
        self._mod_isolated_dancoff_fsr_ids = []
        self._mod_isolated_dancoff_fsr_inds = []

        self._clad_full_dancoff_fsr_ids = []
        self._clad_full_dancoff_fsr_inds = []
        self._mod_full_dancoff_fsr_ids = []
        self._mod_full_dancoff_fsr_inds = []

        # ======================================================================
        # TRANSPORT CALCULATION DATA
        # ----------------------------------------------------------------------
        self._clad_fsr_ids: List[int] = []
        self._mod_fsr_ids: List[int] = []

        self._clad_fsr_inds: List[int] = []
        self._mod_fsr_inds: List[int] = []

        # Holds the CrossSection object used for the real transport calc.
        self._clad_xs: Optional[CrossSection] = None

    @property
    def clad(self) -> Material:
        return self._clad

    @property
    def inner_radius(self) -> float:
        return self._inner_radius

    @property
    def outer_radius(self) -> float:
        return self._outer_radius

    @property
    def fill(self) -> Optional[BurnablePoisonRod]:
        return self._fill

    @property
    def empty(self) -> bool:
        return self.fill is None

    @property
    def clad_dancoff_corrections(self) -> List[float]:
        return self._clad_dancoff_corrections

    def _check_dx_dy(self, dx, dy, pintype):
        if pintype == PinCellType.Full:
            if dx < 2.0 * self.outer_radius:
                raise ValueError(
                    "The fuel pin cell x width must be > the diameter of the cladding."
                )
            if dy < 2.0 * self.outer_radius:
                raise ValueError(
                    "The fuel pin cell y width must be > the diameter of the cladding."
                )
        elif pintype in [PinCellType.XN, PinCellType.XP]:
            if dx < self.outer_radius:
                raise ValueError(
                    "The fuel pin cell x width must be > the radius of the cladding."
                )
            if dy < 2.0 * self.outer_radius:
                raise ValueError(
                    "The fuel pin cell y width must be > the diameter of the cladding."
                )
        elif pintype in [PinCellType.YN, PinCellType.YP]:
            if dy < self.outer_radius:
                raise ValueError(
                    "The fuel pin cell y width must be > the radius of the cladding."
                )
            if dx < 2.0 * self.outer_radius:
                raise ValueError(
                    "The fuel pin cell x width must be > the diameter of the cladding."
                )
        else:
            if dx < self.outer_radius:
                raise ValueError(
                    "The fuel pin cell x width must be > the radius of the cladding."
                )
            if dy < self.outer_radius:
                raise ValueError(
                    "The fuel pin cell y width must be > the radius of the cladding."
                )

    def load_nuclides(self, ndl: NDLibrary) -> None:
        """
        Loads all the nuclides for all current materials into the data library.

        Parameters
        ----------
        ndl : NDLibrary
            Nuclear data library which should load the nuclides.
        """
        self.clad.load_nuclides(ndl)

    # ==========================================================================
    # Dancoff Correction Related Methods
    def set_xs_for_fuel_dancoff_calculation(self) -> None:
        """
        Sets the 1-group cross sections to calculate the fuel Dancoff
        corrections.
        """
        self._clad_dancoff_xs.set(
            CrossSection(
                np.array([self.clad.potential_xs]),
                np.array([self.clad.potential_xs]),
                np.array([[0.0]]),
                "Clad",
            )
        )

        if isinstance(self.fill, BurnablePoisonRod):
            self.fill.set_xs_for_dancoff_calculation()

    def set_xs_for_clad_dancoff_calculation(self, ndl: NDLibrary) -> None:
        """
        Sets the 1-group cross sections to calculate the clad Dancoff
        corrections.

        Parameters
        ----------
        ndl : NDLibrary
            Nuclear data library for obtaining potential scattering cross
            sections.
        """
        self._clad_dancoff_xs.set(
            CrossSection(
                np.array([1.0e5]),
                np.array([1.0e5]),
                np.array([[0.0]]),
                "Clad",
            )
        )

        if isinstance(self.fill, BurnablePoisonRod):
            self.fill.set_xs_for_dancoff_calculation()

    def make_dancoff_moc_cell(
        self,
        moderator_xs: CrossSection,
        dx: float,
        dy: float,
        pintype: PinCellType,
        isolated: bool,
    ) -> SimplePinCell:
        """
        Makes a simplified cell suitable for performing Dancoff correction
        calculations. The flat source region IDs are stored locally in the
        GuideTube object.

        Parameters
        ----------
        moderator_xs : CrossSection
            One group cross sections for the moderator. Total should equal
            absorption (i.e. no scattering) and should be equal to the
            macroscopic potential cross section.
        dx : float
            Width of the cell along x.
        dy : float
            Width of the cell along y.
        pintype : PinCellType
            How the cell should be split (along x, y, or only a quadrant).
        isolated : bool
            If True, the FSR IDs are stored for the isolated pin. Otherwise,
            they are stored for the full pin.

        Returns
        -------
        SimplifiedPinCell
            Pin cell object for MOC Dancoff correction calculation.
        """
        self._check_dx_dy(dx, dy, pintype)

        # First we create list of radii and materials
        radii = []
        xs = []

        if isinstance(self.fill, BurnablePoisonRod):
            radii, xs = self.fill._make_dancoff_moc_cell(moderator_xs)
        FILL_OFFSET = len(radii)

        # This is either all the moderator inside the guide tube, or it is the
        # ring of moderator inside the guide tube, but outside the fill object.
        radii.append(self.inner_radius)
        xs.append(moderator_xs)

        radii.append(self.outer_radius)
        xs.append(self._clad_dancoff_xs)

        xs.append(moderator_xs)

        # Make the simple pin cell.
        cell = SimplePinCell(radii, xs, dx, dy, pintype)

        # Get the FSR IDs for the regions of interest
        cell_fsr_ids = list(cell.get_all_fsr_ids())
        cell_fsr_ids.sort()

        if isolated:
            self._clad_isolated_dancoff_fsr_ids.append(cell_fsr_ids[FILL_OFFSET + 1])
            self._mod_isolated_dancoff_fsr_ids = [
                cell_fsr_ids[FILL_OFFSET + 0],
                cell_fsr_ids[FILL_OFFSET + 2],
            ]

            # Must set FSR IDs by hand for the fill
            if isinstance(self.fill, BurnablePoisonRod):
                if self.fill.center is not None:
                    self._mod_isolated_dancoff_fsr_ids.append(cell_fsr_ids[0])
                else:
                    self.fill._center_isolated_dancoff_fsr_ids.append(cell_fsr_ids[0])

                self.fill._clad_isolated_dancoff_fsr_ids = [
                    cell_fsr_ids[1],
                    cell_fsr_ids[5],
                ]
                self.fill._gap_isolated_dancoff_fsr_ids = [
                    cell_fsr_ids[2],
                    cell_fsr_ids[4],
                ]
                self.fill._poison_isolated_dancoff_fsr_ids = [cell_fsr_ids[3]]

        else:
            self._clad_full_dancoff_fsr_ids.append(cell_fsr_ids[FILL_OFFSET + 1])
            self._mod_full_dancoff_fsr_ids = [
                cell_fsr_ids[FILL_OFFSET + 0],
                cell_fsr_ids[FILL_OFFSET + 2],
            ]

            # Must set FSR IDs by hand for the fill
            if isinstance(self.fill, BurnablePoisonRod):
                if self.fill.center is not None:
                    self._mod_full_dancoff_fsr_ids.append(cell_fsr_ids[0])
                else:
                    self.fill._center_full_dancoff_fsr_ids.append(cell_fsr_ids[0])

                self.fill._clad_full_dancoff_fsr_ids = [
                    cell_fsr_ids[1],
                    cell_fsr_ids[5],
                ]
                self.fill._gap_full_dancoff_fsr_ids = [cell_fsr_ids[2], cell_fsr_ids[4]]
                self.fill._poison_full_dancoff_fsr_ids = [cell_fsr_ids[3]]

        return cell

    def populate_dancoff_fsr_indexes(
        self, isomoc: MOCDriver, fullmoc: MOCDriver
    ) -> None:
        """
        Obtains the flat source region indexes for all of the flat source
        regions used in the Dancoff correction calculations.

        Parameters
        ----------
        isomoc : MOCDriver
            MOC simulation for the isolated pin.
        fullmoc : MOCDriver
            MOC simulation for the full geometry.
        """
        self._clad_isolated_dancoff_fsr_inds = []
        self._mod_isolated_dancoff_fsr_inds = []

        self._clad_full_dancoff_fsr_inds = []
        self._mod_full_dancoff_fsr_inds = []

        for id in self._clad_isolated_dancoff_fsr_ids:
            self._clad_isolated_dancoff_fsr_inds.append(isomoc.get_fsr_indx(id, 0))
        for id in self._mod_isolated_dancoff_fsr_ids:
            self._mod_isolated_dancoff_fsr_inds.append(isomoc.get_fsr_indx(id, 0))

        for id in self._clad_full_dancoff_fsr_ids:
            self._clad_full_dancoff_fsr_inds.append(fullmoc.get_fsr_indx(id, 0))
        for id in self._mod_full_dancoff_fsr_ids:
            self._mod_full_dancoff_fsr_inds.append(fullmoc.get_fsr_indx(id, 0))

        if isinstance(self.fill, BurnablePoisonRod):
            self.fill.populate_dancoff_fsr_indexes(isomoc, fullmoc)

    def set_isolated_dancoff_fuel_sources(
        self, isomoc: MOCDriver, moderator: Material
    ) -> None:
        """
        Initializes the fixed sources for the isolated MOC calculation required
        in computing Dancoff corrections. Sources are set for a fuel Dancoff
        correction calculation.

        Parameters
        ----------
        isomoc : MOCDriver
            MOC simulation for the isolated geometry.
        moderator : Material
            Material definition for the moderator, used to obtain the potential
            scattering cross section.
        """
        # Clad sources should all be potential_xs
        pot_xs = self.clad.potential_xs
        for ind in self._clad_isolated_dancoff_fsr_inds:
            isomoc.set_extern_src(ind, 0, pot_xs)

        # Moderator sources should all be potential_xs
        pot_xs = moderator.potential_xs
        for ind in self._mod_isolated_dancoff_fsr_inds:
            isomoc.set_extern_src(ind, 0, pot_xs)

        if isinstance(self.fill, BurnablePoisonRod):
            self.fill.set_isolated_dancoff_fuel_sources(isomoc, moderator)

    def set_isolated_dancoff_clad_sources(
        self, isomoc: MOCDriver, moderator: Material, ndl: NDLibrary
    ) -> None:
        """
        Initializes the fixed sources for the isolated MOC calculation required
        in computing Dancoff corrections. Sources are set for a clad Dancoff
        correction calculation.

        Parameters
        ----------
        isomoc : MOCDriver
            MOC simulation for the isolated geometry.
        moderator : Material
            Material definition for the moderator, used to obtain the potential
            scattering cross section.
        ndl : NDLibrary
            Nuclear data library for obtaining potential scattering cross
            sections.
        """
        # Clad sources should all be zero !
        for ind in self._clad_isolated_dancoff_fsr_inds:
            isomoc.set_extern_src(ind, 0, 0.0)

        # Moderator sources should all be potential_xs
        pot_xs = moderator.potential_xs
        for ind in self._mod_isolated_dancoff_fsr_inds:
            isomoc.set_extern_src(ind, 0, pot_xs)

        if isinstance(self.fill, BurnablePoisonRod):
            self.fill.set_isolated_dancoff_clad_sources(isomoc, moderator, ndl)

    def set_full_dancoff_fuel_sources(
        self, fullmoc: MOCDriver, moderator: Material
    ) -> None:
        """
        Initializes the fixed sources for the full MOC calculation required
        in computing Dancoff corrections. Sources are set for a fuel Dancoff
        correction calculation.

        Parameters
        ----------
        fullmoc : MOCDriver
            MOC simulation for the full geometry.
        moderator : Material
            Material definition for the moderator, used to obtain the potential
            scattering cross section.
        """
        # Clad sources should all be potential_xs
        pot_xs = self.clad.potential_xs
        for ind in self._clad_full_dancoff_fsr_inds:
            fullmoc.set_extern_src(ind, 0, pot_xs)

        # Moderator sources should all be potential_xs
        pot_xs = moderator.potential_xs
        for ind in self._mod_full_dancoff_fsr_inds:
            fullmoc.set_extern_src(ind, 0, pot_xs)

        if isinstance(self.fill, BurnablePoisonRod):
            self.fill.set_full_dancoff_fuel_sources(fullmoc, moderator)

    def set_full_dancoff_clad_sources(
        self, fullmoc: MOCDriver, moderator: Material, ndl: NDLibrary
    ) -> None:
        """
        Initializes the fixed sources for the full MOC calculation required
        in computing Dancoff corrections. Sources are set for a clad Dancoff
        correction calculation.

        Parameters
        ----------
        isomoc : MOCDriver
            MOC simulation for the isolated geometry.
        moderator : Material
            Material definition for the moderator, used to obtain the potential
            scattering cross section.
        ndl : NDLibrary
            Nuclear data library for obtaining potential scattering cross
            sections.
        """
        # Clad sources should all be zero !
        for ind in self._clad_full_dancoff_fsr_inds:
            fullmoc.set_extern_src(ind, 0, 0.0)

        # Moderator sources should all be potential_xs
        pot_xs = moderator.potential_xs
        for ind in self._mod_full_dancoff_fsr_inds:
            fullmoc.set_extern_src(ind, 0, pot_xs)

        if isinstance(self.fill, BurnablePoisonRod):
            self.fill.set_full_dancoff_clad_sources(fullmoc, moderator, ndl)

    def compute_clad_dancoff_correction(
        self, isomoc: MOCDriver, fullmoc: MOCDriver
    ) -> float:
        """
        Computes the Dancoff correction for the cladding region.

        Parameters
        ----------
        isomoc : MOCDriver
            MOC simulation for the isolated geometry (previously solved).
        fullmoc : MOCDriver
            MOC simulation for the full geometry (previously solved).

        Returns
        -------
        float
            Dancoff correction for the cladding region.
        """
        iso_flux = isomoc.homogenize_flux_spectrum(
            self._clad_isolated_dancoff_fsr_inds
        )[0]
        full_flux = fullmoc.homogenize_flux_spectrum(self._clad_full_dancoff_fsr_inds)[
            0
        ]
        C = (iso_flux - full_flux) / iso_flux
        D = 1.0 - C
        return D

    def append_clad_dancoff_correction(self, C) -> None:
        """
        Saves new Dancoff correction for the cladding that will be used for all
        subsequent cross section updates.

        Parameters
        ----------
        C : float
            New Dancoff correction.
        """
        if C < 0.0 or C > 1.0:
            raise ValueError(
                f"Dancoff correction must be in range [0, 1]. Was provided {C}."
            )
        self._clad_dancoff_corrections.append(C)

    # ==========================================================================
    # Transport Calculation Related Methods
    def set_clad_xs_for_depletion_step(self, t: int, ndl: NDLibrary) -> None:
        """
        Constructs the CrossSection object for the cladding of the guide tube
        at the specified depletion step. The depletion step only changes the
        Dancoff correction, not the cladding composition.

        Parameters
        ----------
        t : int
            Index for the depletion step.
        ndl : NDLibrary
            Nuclear data library to use for cross sections.
        """
        # Compute escape xs
        Ee = 1.0 / (2.0 * (self.outer_radius - self.inner_radius))

        # Get / set the xs
        if self._clad_xs is None:
            self._clad_xs = self.clad.roman_xs(
                self._clad_dancoff_corrections[t], Ee, ndl
            )
        else:
            self._clad_xs.set(
                self.clad.roman_xs(self._clad_dancoff_corrections[t], Ee, ndl)
            )

        if self._clad_xs.name == "":
            self._clad_xs.name = "Clad"

    def set_fill_xs_for_depletion_step(self, t: int, ndl: NDLibrary) -> None:
        """
        Constructs the CrossSection objects for the fill of the guide tube
        at the specified depletion step. The depletion step changes the poison
        composition, if filled with a burnable poison rod.

        Parameters
        ----------
        t : int
            Index for the depletion step.
        ndl : NDLibrary
            Nuclear data library to use for cross sections.
        """
        if self.empty:
            return

        if isinstance(self.fill, BurnablePoisonRod):
            self.fill.set_center_xs(ndl)
            self.fill.set_gap_xs(ndl)
            self.fill.set_clad_xs(ndl)
            self.fill.set_poison_xs_for_depletion_step(t, ndl)

    def make_moc_cell(
        self, moderator_xs: CrossSection, dx: float, dy: float, pintype: PinCellType
    ) -> PinCell:
        """
        Constructs the pin cell object used in for the global MOC simulation.

        Parameters
        ----------
        moderator_xs : CrossSection
            Cross sections to use for the moderator surrounding the fuel pin.
        dx : float
            Width of the cell along x.
        dy : float
            Width of the cell along y.
        pintype : PinCellType
            How the pin cell should be split (along x, y, or only a quadrant).
        """
        if self._clad_xs is None:
            raise RuntimeError("Clad cross section has not yet been built.")
        self._check_dx_dy(dx, dy, pintype)

        # Initialize empty lists
        radii = []
        xss = []
        FILL_OFFSET = 0
        NUM_MOD_RINGS = 3

        if self.empty:
            # For an empty guide tube, we use 3 rings of water
            # Create list of inner moderator radii
            V = np.pi * self.inner_radius * self.inner_radius
            Vr = V / 3.0
            for ri in range(3):
                Rin = 0.0
                if ri > 0:
                    Rin = radii[-1]
                Rout = np.sqrt((Vr + np.pi * Rin * Rin) / np.pi)
                if Rout > self.inner_radius:
                    Rout = self.inner_radius
                radii.append(Rout)

            # Initialize the cross section lists with the inner moderator xs
            xss += len(radii) * [moderator_xs]
        elif isinstance(self.fill, BurnablePoisonRod):
            radii, xss = self.fill._make_moc_cell(moderator_xs)
            FILL_OFFSET = len(radii)

            # Add the layer of moderator outside the poison rod
            radii.append(self.inner_radius)
            xss.append(moderator_xs)
            NUM_MOD_RINGS = 1
        else:
            # Should never get here due to check in constructor, but in any case
            raise TypeError("Unknown fill object placed in guide tube.")

        # Add cladding
        radii.append(self.outer_radius)
        xss.append(self._clad_xs)

        # Add another ring of moderator if possible
        if (
            pintype == PinCellType.Full
            and min(dx, dy) > 2.0 * self.outer_radius
            and 0.5 * min(dx, dy) - self.outer_radius >= 0.1
        ):
            radii.append(0.5 * min(dx, dy))
            xss.append(moderator_xs)
        elif (
            pintype in [PinCellType.XN, PinCellType.XP]
            and dx > self.outer_radius
            and dy > 2.0 * self.outer_radius
            and min(dx, 0.5 * dy) - self.outer_radius >= 0.1
        ):
            radii.append(min(dx, 0.5 * dy))
            xss.append(moderator_xs)
        elif (
            pintype in [PinCellType.YN, PinCellType.YP]
            and dy > self.outer_radius
            and dx > 2.0 * self.outer_radius
            and min(0.5 * dx, dy) - self.outer_radius >= 0.1
        ):
            radii.append(min(0.5 * dx, dy))
            xss.append(moderator_xs)
        elif (
            pintype in [PinCellType.I, PinCellType.II, PinCellType.III, PinCellType.IV]
            and dx > self.outer_radius
            and dy > self.outer_radius
            and min(dx, dy) - self.outer_radius >= 0.1
        ):
            radii.append(min(dx, dy))
            xss.append(moderator_xs)

        # Add moderator to the end of materials
        xss.append(moderator_xs)

        cell = PinCell(radii, xss, dx, dy, pintype)

        # Get the FSR IDs for the regions of interest
        cell_fsr_ids = list(cell.get_all_fsr_ids())
        cell_fsr_ids.sort()

        # Number of angular divisions
        NA = 8
        if pintype in [PinCellType.XN, PinCellType.XP, PinCellType.YN, PinCellType.YP]:
            NA = 4
        elif pintype in [
            PinCellType.I,
            PinCellType.II,
            PinCellType.III,
            PinCellType.IV,
        ]:
            NA = 2

        # Get all the FSR ids for the poison rod
        if isinstance(self.fill, BurnablePoisonRod):
            if self.fill.center is not None:
                self.fill._center_fsr_ids += cell_fsr_ids[0:NA]
            self.fill._clad_fsr_ids += cell_fsr_ids[NA : 2 * NA]
            self.fill._gap_fsr_ids += cell_fsr_ids[2 * NA : 3 * NA]
            self.fill._poison_fsr_ids += cell_fsr_ids[3 * NA : 4 * NA]
            self.fill._gap_fsr_ids += cell_fsr_ids[4 * NA : 5 * NA]
            self.fill._clad_fsr_ids += cell_fsr_ids[5 * NA : 6 * NA]

        I = 0  # Starting index for cell_fsr_inds
        # Go through all rings of moderator and get FSR IDs
        for a in range(NUM_MOD_RINGS * NA):
            self._mod_fsr_ids.append(cell_fsr_ids[FILL_OFFSET * NA + I])
            I += 1

        # If we have a poision rod with a water center, add those FSR IDs to the moderator list
        if isinstance(self.fill, BurnablePoisonRod):
            if self.fill.center is None:
                self._mod_fsr_ids += cell_fsr_ids[0:NA]

        # Get FSR IDs for the cladding
        for a in range(NA):
            self._clad_fsr_ids.append(cell_fsr_ids[FILL_OFFSET * NA + I])
            I += 1

        # Everything outside the clad should be moderator
        self._mod_fsr_ids = list(cell_fsr_ids[FILL_OFFSET * NA + I :])

        return cell

    def populate_fsr_indexes(self, moc: MOCDriver) -> None:
        """
        Obtains the flat source region indexes for all of the flat source
        regions used in the full MOC calculations.

        Parameters
        ----------
        moc : MOCDriver
            MOC simulation for the full calculations.
        """
        self._clad_fsr_inds: List[int] = []
        self._mod_fsr_inds: List[int] = []

        for id in self._clad_fsr_ids:
            self._clad_fsr_inds.append(moc.get_fsr_indx(id, 0))
        for id in self._mod_fsr_ids:
            self._mod_fsr_inds.append(moc.get_fsr_indx(id, 0))

        if isinstance(self.fill, BurnablePoisonRod):
            self.fill.populate_fsr_indexes(moc)

    def obtain_flux_spectra(self, moc: MOCDriver) -> None:
        """
        If the guide tube contains a burnable poison rod, the average flux
        spectrum in the poison is obtained from the MOC simulation.

        Parameters
        ----------
        moc : MOCDriver
            MOC simulation for the full calculations.
        """
        if not self.empty and isinstance(self.fill, BurnablePoisonRod):
            self.fill.obtain_flux_spectra(moc)

    def normalize_flux_spectrum(self, f) -> None:
        """
        If the guide tube contains a burnable poison rod, it applies a
        multiplicative factor to the flux spectra for the poison. This permits
        normalizing the flux to a known assembly power.

        Parameters
        ----------
        f : float
            Normalization factor.
        """
        if not self.empty and isinstance(self.fill, BurnablePoisonRod):
            self.fill.normalize_flux_spectrum(f)

    def predict_depletion(
        self,
        chain: DepletionChain,
        ndl: NDLibrary,
        dt: float,
        dtm1: Optional[float] = None,
    ) -> None:
        """
        Performs the predictor in the integration of the Bateman equation.
        If the argument for the previous time step is not provided, CE/LI will
        be used. Otherwise, CE/LI is used on the first depletion step, and
        LE/QI is used for all subsequent time steps. The predicted material
        compositions are appended to the materials lists.

        Paramters
        ---------

        chain : DepletionChain
            Depletion chain to use for radioactive decay and transmutation.
        ndl : NDLibrary
            Nuclear data library.
        dt : float
            Durration of the time step in seconds.
        dtm1 : float, optional
            Durration of the previous time step in seconds. Default is None.
        """
        if dt <= 0:
            raise ValueError("Predictor time step must be > 0.")

        if not self.empty and isinstance(self.fill, BurnablePoisonRod):
            self.fill.predict_depletion(chain, ndl, dt, dtm1)

    def correct_depletion(
        self,
        chain: DepletionChain,
        ndl: NDLibrary,
        dt: float,
        dtm1: Optional[float] = None,
    ) -> None:
        """
        Performs the corrector in the integration of the Bateman equation.
        If the argument for the previous time step is not provided, CE/LI will
        be used. Otherwise, CE/LI is used on the first depletion step, and
        LE/QI is used for all subsequent time steps. The corrected material
        compositions replace the ones where were appended in the corrector step.

        Parameters
        ----------
        chain : DepletionChain
            Depletion chain to use for radioactive decay and transmutation.
        ndl : NDLibrary
            Nuclear data library.
        dt : float
            Durration of the time step in seconds.
        dtm1 : float, optional
            Durration of the previous time step in seconds. Default is None.
        """
        if dt <= 0:
            raise ValueError("Corrector time step must be > 0.")
        # Nothing to do here yet, as guide tube "fills" with burnable
        # absorber pins is not yet supported. In the future, those will
        # need to be depleted !

        if not self.empty and isinstance(self.fill, BurnablePoisonRod):
            self.fill.correct_depletion(chain, ndl, dt, dtm1)



import pytest
import pytest_check as check
from scarabee import CMFD, CMFDSurfaceCrossing, CMFDSurfaceCrossingType, Direction, Vector
"""
Unit tests for CMFD::get_surface() method

The test mesh consists of a 2x2 rectangular grid with dx, dy = 1.0,1.0
"""
class TestCMFDSurfaces:
    @pytest.fixture
    
    def make_cmfd(self):
        """
        Makes the test grid. Cell indexes are:
        [2, 3]
        [0, 1]
        """
        cmfd = CMFD([1.0,1.0],[1.0,1.0],[(0,1)])
        return cmfd
    
    def test_cell_XP_to_XN(self, make_cmfd):
        """
        Tests get_surface() for a point, direction pair that goes from 
        the positive x surface of a cell into the negative x surface of 
        an adjacent cell.

        Should return a valid crossing for the negative x surface of cell 1
        """
        cmfd = make_cmfd
        tol = 1e-12
        r = Vector(0.0-tol,-0.5)
        u = Direction(1.0,0.0)

        surface = cmfd.get_surface(r,u)

        check.equal(surface.is_valid, True)
        check.equal(surface.cell_index, 1)
        check.equal(surface.crossing, CMFDSurfaceCrossingType.XN)

    def test_XN_to_outside_left(self, make_cmfd):
        """
        Tests get_surface() for a point, direction pair that goes 
        from the left of a cell to an exterior boundary
        Should throw an error
        """
        cmfd = make_cmfd
        tol = 1e-12
        r = Vector(-1.0+tol,-0.5)
        u = Direction(-1.0,0.0)
        with pytest.raises(RuntimeError):
            surface = cmfd.get_surface(r,u)

    def test_YP_to_YN(self, make_cmfd):
        """
        Tests get_surface() for a point, direction pair that goes from 
        the positive y surface of a cell into the negative y surface of 
        an adjacent cell.

        Should return a valid crossing for the negative y surface of cell 2
        """
        cmfd = make_cmfd
        tol = 1e-12
        r = Vector(-0.5,0.0-tol)
        u = Direction(0.0,1.0)

        surface = cmfd.get_surface(r,u)

        check.equal(surface.is_valid, True)
        check.equal(surface.cell_index, 2)
        check.equal(surface.crossing, CMFDSurfaceCrossingType.YN)

    def test_YN_to_outside_bottom(self, make_cmfd):
        """
        Tests get_surface() for a point, direction pair that goes 
        from the bottom of a cell to an exterior boundary
        Should throw an error
        """
        cmfd = make_cmfd
        tol = 1e-12
        r = Vector(-0.5,-1.0+tol)
        u = Direction(0,-1)

        with pytest.raises(RuntimeError):
            surface = cmfd.get_surface(r,u)


    def test_I_to_III(self, make_cmfd):
        """
        Tests get_surface() for a point, direction pair that goes from 
        the top right surface of a cell into the bottom left surface of 
        an adjacent cell.

        Should return a valid crossing for the III surface of cell 3 
        """
        cmfd = make_cmfd
        tol = 1e-12
        r = Vector(0.0-tol,0.0-tol)
        u = Direction(0.707,0.707)

        surface = cmfd.get_surface(r,u)

        check.equal(surface.is_valid, True)
        check.equal(surface.cell_index, 3)
        check.equal(surface.crossing, CMFDSurfaceCrossingType.III)

    def test_II_to_IV(self, make_cmfd):
        """
        Tests get_surface() for a point, direction pair that goes from 
        the top left surface of a cell into the bottom right surface of 
        an adjacent cell.

        Should return a valid crossing for the IV surface of cell 2
        """
        cmfd = make_cmfd
        tol = 1e-12
        r = Vector(0.0+tol,0.0-tol)
        u = Direction(-0.707,0.707)

        surface = cmfd.get_surface(r,u)

        check.equal(surface.is_valid, True)
        check.equal(surface.cell_index, 2)
        check.equal(surface.crossing, CMFDSurfaceCrossingType.IV)

    def test_IV_to_II(self, make_cmfd):
        """
        Tests get_surface() for a point, direction pair that goes from 
        the bottom right surface of a cell into the top left surface of 
        an adjacent cell.

        Should return a valid crossing for the top left surface of cell 1
        """
        cmfd = make_cmfd
        tol = 1e-12
        r = Vector(0.0-tol,0.0+tol)
        u = Direction(0.707,-0.707)

        surface = cmfd.get_surface(r,u)

        check.equal(surface.is_valid, True)
        check.equal(surface.cell_index, 1)
        check.equal(surface.crossing, CMFDSurfaceCrossingType.II)

    def test_III_to_I(self, make_cmfd):
        """
        Tests get_surface() for a point, direction pair that goes from 
        the bottom left surface of a cell into the top right surface of 
        an adjacent cell.

        Should return a valid crossing for the top right surface of cell 0
        """
        cmfd = make_cmfd
        tol = 1e-12
        r = Vector(0.0+tol,0.0+tol)
        u = Direction(-0.707,-0.707)

        surface = cmfd.get_surface(r,u)

        check.equal(surface.is_valid, True)
        check.equal(surface.cell_index, 0)
        check.equal(surface.crossing, CMFDSurfaceCrossingType.I)







# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))

# Check if we are on Read the Docs server
on_rtd = os.environ.get('READTHEDOCS', None) == 'True'

# -- Project information -----------------------------------------------------

project = 'Scarabée'
copyright = '2024-2025, Hunter Belanger'
author = 'Hunter Belanger'

# The full version, including alpha/beta/rc tags
release = '0.1.0'

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
  'sphinx.ext.autodoc',
  'sphinx.ext.intersphinx',
  'sphinx.ext.autosummary',
  'sphinx.ext.napoleon',
  'sphinx.ext.autosectionlabel',
  'sphinx.ext.todo',
  'sphinx.ext.coverage',
  'sphinx.ext.mathjax',
  'sphinx.ext.ifconfig',
  'sphinx.ext.viewcode',
  'nbsphinx'
]

# Helps reduce duplicate label warnings from autosectionlabel
autosectionlabel_prefix_document = True

autosectionlabel_maxdepth = 4

#suppress_warnings = [
#    'autosectionlabel.*',
#]

autodoc_default_options = {
    'members': True,
    'member-order': 'bysource',
    'special-members': '__init__',
    'inherited-members': True,
}

# Napoleon settings
napoleon_google_docstring = True
napoleon_numpy_docstring = True
napoleon_include_init_with_doc = False
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = True
napoleon_use_admonition_for_examples = False
napoleon_use_admonition_for_notes = False
napoleon_use_admonition_for_references = False
napoleon_use_ivar = False
napoleon_use_param = True
napoleon_use_rtype = True
napoleon_preprocess_types = False
napoleon_type_aliases = None
napoleon_attr_annotations = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

numfig = True

# Mathjax Configuration
mathjax3_config = {
  'loader': {'load': ['[tex]/newcommand', '[tex]/physics', '[tex]/boldsymbol']},
  'tex': {
    'packages': {'[+]': ['newcommand', 'physics', 'boldsymbol']},
    'macros': {
      'pos': '\\boldsymbol{r}',
      'dir': '\\boldsymbol{\hat{\Omega}}',
      'Et': '\\Sigma_t',
      'Es': '\\Sigma_s',
      'Ea': '\\Sigma_a',
      'Ef': '\\Sigma_f',
      'keff': 'k_\\text{eff}'
    }
  }
}

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#import sphinx_rtd_theme
#html_theme = 'sphinx_rtd_theme'
#html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
#html_static_path = ['_static']


# Seetings for the PyData theme
html_theme = 'pydata_sphinx_theme'

#html_title = 'Scarabée'
#html_logo = "_images/scarabee_logo.png"
html_favicon = "_images/scarabee_logo.png"

html_theme_options = {
   "icon_links": [
        {
            # Label for this link
            "name": "GitHub",
            # URL where the link will redirect
            "url": "https://github.com/scarabee-dev/scarabee",  # required
            # Icon class (if "type": "fontawesome"), or path to local image (if "type": "local")
            "icon": "fab fa-github-square",
            # The type of image to be used (see below for details)
            "type": "fontawesome",
        }
   ],

   "logo": {
      "text": "Scarabée",
      "image_light": "_images/scarabee_logo.png",
      "image_dark": "_images/scarabee_logo_dark.png",
   }
}



from scarabee import *
import numpy as np
import matplotlib.pyplot as plt

ndl = NDLibrary()

def get_fuel_dancoff_correction(Zirc, Water):
  Et = np.array([1.E5])
  Ea = np.array([1.E5])
  Es = np.array([[0.]])
  Fuel = CrossSection(Et, Ea, Es, "Fuel")

  Et = np.array([Zirc.potential_xs])
  Ea = np.array([Zirc.potential_xs])
  Clad = CrossSection(Et, Ea, Es, "Clad")

  Et = np.array([Water.potential_xs])
  Ea = np.array([Water.potential_xs])
  Mod = CrossSection(Et, Ea, Es, "Mod")

  # Define Cells
  pitch = 1.26
  iso_pitch = 50.*pitch
  fuel_rad = 0.4095
  clad_rad = 0.475

  # Simulate isolated geometry
  radii = [fuel_rad, clad_rad]
  mats =  [Fuel, Clad, Mod]
  isolated_pin = SimplePinCell(radii, mats, iso_pitch, iso_pitch)
  iso_c2d = Cartesian2D([iso_pitch], [iso_pitch])
  iso_c2d.set_tiles([isolated_pin])
  iso_moc = MOCDriver(iso_c2d)
  for i in range(iso_moc.nfsr):
    i_xs = iso_moc.xs(i)
    if i_xs.name == "Fuel":
      iso_moc.set_extern_src(i, 0, 0.)
    elif i_xs.name == "Clad":
      iso_moc.set_extern_src(i, 0, Zirc.potential_xs)
    else:
      iso_moc.set_extern_src(i, 0, Water.potential_xs)
  iso_moc.sim_mode = SimulationMode.FixedSource
  iso_moc.generate_tracks(128, 0.01, YamamotoTabuchi6())
  iso_moc.flux_tolerance = 1.E-5
  iso_moc.solve()

  # Simulate real geomemtry
  radii = [fuel_rad, clad_rad]
  mats =  [Fuel, Clad, Mod]
  P = SimplePinCell(radii, mats, pitch, pitch)
  c2d = Cartesian2D([pitch], [pitch])
  c2d.set_tiles([P])
  moc = MOCDriver(c2d)
  # Set the source
  for i in range(moc.nfsr):
    i_xs = moc.xs(i)
    if i_xs.name == "Fuel":
      moc.set_extern_src(i, 0, 0.)
    elif i_xs.name == "Clad":
      moc.set_extern_src(i, 0, Zirc.potential_xs)
    else:
      moc.set_extern_src(i, 0, Water.potential_xs)
  moc.generate_tracks(128, 0.01, YamamotoTabuchi6())
  moc.flux_tolerance = 1.E-5
  moc.sim_mode = SimulationMode.FixedSource
  moc.solve()

  # Calculate Dancoff correction factor
  iso_flux = iso_moc.flux(Vector(0.,0.), Direction(1.,0.), 0)
  flux = moc.flux(Vector(0.,0.), Direction(1.,0.), 0)
  C = (iso_flux - flux) / iso_flux
  return C

def get_clad_dancoff_correction(UO2, Water):
  Et = np.array([UO2.potential_xs])
  Ea = np.array([])
  Es = np.array([[0.]])
  Fuel = CrossSection(Et, Ea, Es, "Fuel")

  Et = np.array([1.E5])
  Ea = np.array([1.E5])
  Clad = CrossSection(Et, Ea, Es, "Clad")

  Et = np.array([Water.potential_xs])
  Ea = np.array([Water.potential_xs])
  Mod = CrossSection(Et, Ea, Es, "Mod")

  # Define Cells
  pitch = 1.26
  iso_pitch = 50.*pitch
  fuel_rad = 0.4095
  clad_rad = 0.475

  # Simulate isolated geometry
  radii = [fuel_rad, clad_rad]
  mats =  [Fuel, Clad, Mod]
  isolated_pin = SimplePinCell(radii, mats, iso_pitch, iso_pitch)
  iso_c2d = Cartesian2D([iso_pitch], [iso_pitch])
  iso_c2d.set_tiles([isolated_pin])
  iso_moc = MOCDriver(iso_c2d)
  for i in range(iso_moc.nfsr):
    i_xs = iso_moc.xs(i)
    if i_xs.name == "Fuel":
      iso_moc.set_extern_src(i, 0, UO2.potential_xs)
    elif i_xs.name == "Clad":
      iso_moc.set_extern_src(i, 0, 0.)
    else:
      iso_moc.set_extern_src(i, 0, Water.potential_xs)
  iso_moc.sim_mode = SimulationMode.FixedSource
  iso_moc.generate_tracks(128, 0.01, YamamotoTabuchi6())
  iso_moc.flux_tolerance = 1.E-5
  iso_moc.solve()

  # Simulate real geomemtry
  radii = [fuel_rad, clad_rad]
  mats =  [Fuel, Clad, Mod]
  P = SimplePinCell(radii, mats, pitch, pitch)
  c2d = Cartesian2D([pitch], [pitch])
  c2d.set_tiles([P])
  moc = MOCDriver(c2d)
  # Set the source
  for i in range(moc.nfsr):
    i_xs = moc.xs(i)
    if i_xs.name == "Fuel":
      moc.set_extern_src(i, 0, UO2.potential_xs)
    elif i_xs.name == "Clad":
      moc.set_extern_src(i, 0, 0.)
    else:
      moc.set_extern_src(i, 0, Water.potential_xs)
  moc.generate_tracks(128, 0.01, YamamotoTabuchi6())
  moc.flux_tolerance = 1.E-5
  moc.sim_mode = SimulationMode.FixedSource
  moc.solve()

  # Calculate Dancoff correction factor
  iso_flux = iso_moc.flux(Vector(fuel_rad+0.00001,0.), Direction(1.,0.), 0)
  flux = moc.flux(Vector(fuel_rad+0.00001,0.), Direction(1.,0.), 0)
  C = (iso_flux - flux) / iso_flux
  return C

UO2comp = MaterialComposition()
UO2comp.add_nuclide("U235", 7.0803E-4)
UO2comp.add_nuclide("U238", 2.2604E-2)
UO2comp.add_nuclide("O16",  4.6624E-2)
UO2 = Material(UO2comp, 293.6, ndl)

Zirccomp = MaterialComposition()
Zirccomp.add_nuclide("Zr90", 2.2200E-2)
Zirccomp.add_nuclide("Zr91", 4.8280E-3)
Zirccomp.add_nuclide("Zr92", 7.3713E-3)
Zirccomp.add_nuclide("Zr94", 7.5006E-3)
Zirccomp.add_nuclide("Zr96", 1.2070E-3)
Zirc = Material(Zirccomp, 293.6, ndl)

Watercomp = MaterialComposition()
Watercomp.add_nuclide("H1_H2O", 6.6630E-2)
Watercomp.add_nuclide("O16",    3.3315E-2)
Water = Material(Watercomp, 293.6, ndl)

# Define Cells
pitch = 1.26
fuel_rad = 0.4095
clad_rad = 0.475

set_logging_level(LogLevel.Error)
C_fuel = get_fuel_dancoff_correction(Zirc, Water)
print("Dancoff Correction of Fuel: {}".format(C_fuel))
C_clad = get_fuel_dancoff_correction(UO2, Water)
print("Dancoff Correction of Clad: {}".format(C_clad))
set_logging_level(LogLevel.Info)


# Create fuel xs
Ee = 1. / (2. * fuel_rad)
Fuel = UO2.carlvik_xs(C_fuel, Ee, ndl)

# Create cladding xs using a reference dilution
Ee = 1. / (2. * (clad_rad - fuel_rad))
Clad = Zirc.roman_xs(C_clad, Ee, ndl)

# Create moderator xs using a somewhat random dilution (shouldn't matter except a little bit on O16)
Mod = Water.dilution_xs(Water.size*[1.E10], ndl)

# Solve problem with CP
radii = [0.5*fuel_rad, 0.8*fuel_rad, fuel_rad, clad_rad, 1.2*clad_rad, pitch/np.sqrt(np.pi)]
mats  = [Fuel,         Fuel,         Fuel,     Clad,     Mod,         Mod]
cell = CylindricalCell(radii, mats)
cell.solve()
cell_flux = CylindricalFluxSolver(cell)
cell_flux.flux_tolerance = 1.E-5
cell_flux.solve()

# Get average flux from CP
avg_flux_cp = cell_flux.homogenize_flux_spectrum()
avg_flux_cp = np.flip(avg_flux_cp)

# Create and run full real problem with MOC
radii = [0.5*fuel_rad, 0.8*fuel_rad, fuel_rad, clad_rad, 1.2*clad_rad]
mats  = [Fuel,         Fuel,         Fuel,     Clad,     Mod,         Mod]
P = PinCell(radii, mats, pitch, pitch)
c2d = Cartesian2D([pitch], [pitch])
c2d.set_tiles([P])
moc = MOCDriver(c2d)
moc.generate_tracks(128, 0.01, YamamotoTabuchi6())
moc.flux_tolerance = 1.E-5
moc.solve()

# Get average flux from MOC
avg_flux_moc = moc.homogenize_flux_spectrum()
avg_flux_moc = np.flip(avg_flux_moc)

ebounds = ndl.group_bounds
ebounds = np.flip(ebounds)

# Now we convert to flux per unit lethargy, and compute the lethargy integral
# of each for normalization of each spectrum.
integral_cp = 0.
integral_moc = 0.
for g in range(len(avg_flux_cp)):
    integral_cp += avg_flux_cp[g]
    integral_moc += avg_flux_moc[g]
    Emin = ebounds[g]
    Emax = ebounds[g+1]
    u = np.log(Emax/Emin)
    avg_flux_cp[g] /= u
    avg_flux_moc[g] /= u
avg_flux_cp /= integral_cp
avg_flux_moc /= integral_moc

plt.stairs(avg_flux_cp, edges=ebounds, label="CP")
plt.stairs(avg_flux_moc, edges=ebounds, label="MOC")
plt.xscale('log')
plt.xlabel("Energy [eV]")
plt.ylabel("Flux per Unit Lethargy [Arb. Units]")
plt.legend().set_draggable(True)
plt.show()

#np.save("cp_spectrum.npy", avg_flux_cp)
#np.save("moc_spectrum.npy", avg_flux_moc)
#np.save("energy_bounds.npy", ebounds)
